# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from unittest import TestCase

from grpc import ServicerContext
from mock import Mock, patch
from pyqrllib.pyqrllib import bin2hstr, QRLHelper

from qrl.core.ChainManager import ChainManager
from qrl.core.txs.Transaction import Transaction
from qrl.core.State import State
from qrl.core.misc import logger
from qrl.core.node import POW
from qrl.core.p2p.p2pfactory import P2PFactory
from qrl.core.qrlnode import QRLNode
from qrl.crypto.misc import sha256
from qrl.generated import qrl_pb2
from qrl.services.PublicAPIService import PublicAPIService
from tests.misc.helper import get_alice_xmss, get_bob_xmss, set_default_balance_size, set_qrl_dir, replacement_getTime

logger.initialize_default()


@patch('qrl.core.misc.ntp.getTime', new=replacement_getTime)
class TestPublicAPI(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestPublicAPI, self).__init__(*args, **kwargs)

    @set_default_balance_size()
    def test_messageTxn_sign(self):
        with set_qrl_dir('wallet_ver1'):
            with State() as db_state:
                p2p_factory = Mock(spec=P2PFactory)
                p2p_factory.pow = Mock(spec=POW)
                chain_manager = ChainManager(db_state)

                qrlnode = QRLNode(mining_address=b'')
                qrlnode.set_chain_manager(chain_manager)
                qrlnode._p2pfactory = p2p_factory
                qrlnode._pow = p2p_factory.pow
                qrlnode._peer_addresses = ['127.0.0.1', '192.168.1.1']

                service = PublicAPIService(qrlnode)

                context = Mock(spec=ServicerContext)

                alice = get_alice_xmss()
                my_message = b'Hello QRL!'

                request = qrl_pb2.MessageTxnReq(
                    message=my_message,
                    fee=12,
                    xmss_pk=alice.pk
                )

                response = service.GetMessageTxn(request=request, context=context)
                context.set_code.assert_not_called()
                context.set_details.assert_not_called()

                self.assertIsNotNone(response)
                self.assertIsNotNone(response.extended_transaction_unsigned.tx)
                self.assertEqual('message', response.extended_transaction_unsigned.tx.WhichOneof('transactionType'))

                self.assertEqual(12, response.extended_transaction_unsigned.tx.fee)
                self.assertEqual(alice.pk, response.extended_transaction_unsigned.tx.public_key)
                self.assertEqual(0, response.extended_transaction_unsigned.tx.nonce)
                self.assertEqual(b'', response.extended_transaction_unsigned.tx.signature)
                self.assertEqual(b'', response.extended_transaction_unsigned.tx.transaction_hash)
                self.assertEqual(my_message, response.extended_transaction_unsigned.tx.message.message_hash)

                tmp_hash_pre = response.extended_transaction_unsigned.tx.master_addr
                tmp_hash_pre += response.extended_transaction_unsigned.tx.fee.to_bytes(8, byteorder='big', signed=False)
                tmp_hash_pre += response.extended_transaction_unsigned.tx.message.message_hash

                tmp_hash = sha256(tmp_hash_pre)
                hash_found = bin2hstr(Transaction.from_pbdata(response.extended_transaction_unsigned.tx).
                                      get_data_hash())
                self.assertEqual(hash_found,
                                 bin2hstr(tmp_hash))

                signed_transaction = response.extended_transaction_unsigned.tx
                signed_transaction.signature = alice.sign(tmp_hash)

                req_push = qrl_pb2.PushTransactionReq(transaction_signed=signed_transaction)

                resp_push = service.PushTransaction(req_push, context=context)
                context.set_code.assert_not_called()
                context.set_details.assert_not_called()

                self.assertIsNotNone(resp_push)
                self.assertEqual(qrl_pb2.PushTransactionResp.SUBMITTED,
                                 resp_push.error_code)
                self.assertEqual('b7ee814a548a6bbb8d97b2d3a0eb9e1f8b6ceee49b764e3c7b23d104aca6abeb',
                                 bin2hstr(resp_push.tx_hash))

    @set_default_balance_size()
    def test_transferCoins_get_unsigned(self):
        with set_qrl_dir('wallet_ver1'):
            with State() as db_state:
                p2p_factory = Mock(spec=P2PFactory)
                p2p_factory.pow = Mock(spec=POW)
                chain_manager = ChainManager(db_state)

                qrlnode = QRLNode(mining_address=b'')
                qrlnode.set_chain_manager(chain_manager)
                qrlnode._p2pfactory = p2p_factory
                qrlnode._pow = p2p_factory.pow
                qrlnode._peer_addresses = ['127.0.0.1', '192.168.1.1']

                service = PublicAPIService(qrlnode)

                context = Mock(spec=ServicerContext)

                alice = get_alice_xmss()
                bob = get_bob_xmss()

                request = qrl_pb2.TransferCoinsReq(
                    addresses_to=[bob.address],
                    amounts=[101],
                    fee=12,
                    xmss_pk=alice.pk
                )

                response = service.TransferCoins(request=request, context=context)
                context.set_code.assert_not_called()
                context.set_details.assert_not_called()

                self.assertIsNotNone(response)
                self.assertIsNotNone(response.extended_transaction_unsigned)
                self.assertEqual('transfer',
                                 response.extended_transaction_unsigned.tx.WhichOneof('transactionType'))

                self.assertEqual(12, response.extended_transaction_unsigned.tx.fee)
                self.assertEqual(alice.pk, response.extended_transaction_unsigned.tx.public_key)
                self.assertEqual(0, response.extended_transaction_unsigned.tx.nonce)

                self.assertEqual(b'', response.extended_transaction_unsigned.tx.signature)
                self.assertEqual(b'', response.extended_transaction_unsigned.tx.transaction_hash)

                self.assertEqual(bob.address, response.extended_transaction_unsigned.tx.transfer.addrs_to[0])
                self.assertEqual(101, response.extended_transaction_unsigned.tx.transfer.amounts[0])

    @set_default_balance_size()
    def test_transferCoins_push_unsigned(self):
        with set_qrl_dir('wallet_ver1'):
            with State() as db_state:
                p2p_factory = Mock(spec=P2PFactory)
                p2p_factory.pow = Mock(spec=POW)
                chain_manager = ChainManager(db_state)

                qrlnode = QRLNode(mining_address=b'')
                qrlnode.set_chain_manager(chain_manager)
                qrlnode._p2pfactory = p2p_factory
                qrlnode._pow = p2p_factory.pow
                qrlnode._peer_addresses = ['127.0.0.1', '192.168.1.1']

                service = PublicAPIService(qrlnode)

                context = Mock(spec=ServicerContext)

                alice = get_alice_xmss()
                bob = get_bob_xmss()

                request = qrl_pb2.TransferCoinsReq(
                    addresses_to=[bob.address],
                    amounts=[101],
                    fee=12,
                    xmss_pk=alice.pk
                )

                response = service.TransferCoins(request=request, context=context)
                context.set_code.assert_not_called()
                context.set_details.assert_not_called()

                self.assertIsNotNone(response)
                self.assertIsNotNone(response.extended_transaction_unsigned)
                self.assertEqual('transfer', response.extended_transaction_unsigned.tx.WhichOneof('transactionType'))

                self.assertEqual(12, response.extended_transaction_unsigned.tx.fee)
                self.assertEqual(alice.pk, response.extended_transaction_unsigned.tx.public_key)
                self.assertEqual(0, response.extended_transaction_unsigned.tx.nonce)
                self.assertEqual(b'', response.extended_transaction_unsigned.tx.signature)
                self.assertEqual(b'', response.extended_transaction_unsigned.tx.transaction_hash)
                self.assertEqual(bob.address, response.extended_transaction_unsigned.tx.transfer.addrs_to[0])
                self.assertEqual(101, response.extended_transaction_unsigned.tx.transfer.amounts[0])

                req_push = qrl_pb2.PushTransactionReq(transaction_signed=response.extended_transaction_unsigned.tx)

                resp_push = service.PushTransaction(req_push, context=context)
                context.set_code.assert_not_called()
                context.set_details.assert_not_called()

                self.assertIsNotNone(resp_push)
                self.assertEqual(qrl_pb2.PushTransactionResp.VALIDATION_FAILED,
                                 resp_push.error_code)

    @set_default_balance_size()
    def test_transferCoins_sign(self):
        with set_qrl_dir('wallet_ver1'):
            with State() as db_state:
                p2p_factory = Mock(spec=P2PFactory)
                p2p_factory.pow = Mock(spec=POW)
                chain_manager = ChainManager(db_state)

                qrlnode = QRLNode(mining_address=b'')
                qrlnode.set_chain_manager(chain_manager)
                qrlnode._p2pfactory = p2p_factory
                qrlnode._pow = p2p_factory.pow
                qrlnode._peer_addresses = ['127.0.0.1', '192.168.1.1']

                service = PublicAPIService(qrlnode)

                context = Mock(spec=ServicerContext)

                alice = get_alice_xmss()
                bob = get_bob_xmss()

                request = qrl_pb2.TransferCoinsReq(
                    addresses_to=[bob.address],
                    amounts=[101],
                    fee=12,
                    xmss_pk=alice.pk
                )

                response = service.TransferCoins(request=request, context=context)
                context.set_code.assert_not_called()
                context.set_details.assert_not_called()

                self.assertIsNotNone(response)
                self.assertIsNotNone(response.extended_transaction_unsigned.tx)
                self.assertEqual('transfer', response.extended_transaction_unsigned.tx.WhichOneof('transactionType'))

                self.assertEqual(12, response.extended_transaction_unsigned.tx.fee)
                self.assertEqual(alice.pk, response.extended_transaction_unsigned.tx.public_key)
                self.assertEqual(0, response.extended_transaction_unsigned.tx.nonce)
                self.assertEqual(b'', response.extended_transaction_unsigned.tx.signature)
                self.assertEqual(b'', response.extended_transaction_unsigned.tx.transaction_hash)
                self.assertEqual(bob.address, response.extended_transaction_unsigned.tx.transfer.addrs_to[0])
                self.assertEqual(101, response.extended_transaction_unsigned.tx.transfer.amounts[0])

                tmp_hash_pre = bytes(QRLHelper.getAddress(response.extended_transaction_unsigned.tx.public_key))
                tmp_hash_pre += str(response.extended_transaction_unsigned.tx.fee).encode()
                tmp_hash_pre += response.extended_transaction_unsigned.tx.transfer.addrs_to[0]
                tmp_hash_pre += str(response.extended_transaction_unsigned.tx.transfer.amounts[0]).encode()

                self.assertEqual('010300a1da274e68c88b0ccf448e0b1916fa789b01eb2ed4e9ad565ce264c939078'
                                 '2a9c61ac02f31320103001d65d7e59aed5efbeae64246e0f3184d7c42411421eb38'
                                 '5ba30f2c1c005a85ebc4419cfd313031',
                                 bin2hstr(tmp_hash_pre))

                tmp_hash = sha256(tmp_hash_pre)

                self.assertEqual('3645f2819aba65479f9a7fad3f5d7a41a9357410a595fa02fb947bfe3ed96e0f',
                                 bin2hstr(tmp_hash))

                signed_transaction = response.extended_transaction_unsigned.tx
                signed_transaction.signature = alice.sign(tmp_hash)

                req_push = qrl_pb2.PushTransactionReq(transaction_signed=signed_transaction)

                resp_push = service.PushTransaction(req_push, context=context)
                context.set_code.assert_not_called()
                context.set_details.assert_not_called()

                self.assertIsNotNone(resp_push)
                self.assertEqual(qrl_pb2.PushTransactionResp.SUBMITTED,
                                 resp_push.error_code)
                self.assertEqual('30955fdc5e2d9dbe5fb9bf812f2e1b6c4b409a8a7c7a75f1c3e9ba1ffdd8e60e',
                                 bin2hstr(resp_push.tx_hash))



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
import datetime
import decimal
from decimal import Decimal

from qrl.core import config

START_DATE = datetime.datetime(2018, 4, 1, 0, 0, 0)
END_DATE = datetime.datetime(2218, 4, 1, 0, 0, 0)
c = END_DATE - START_DATE
TOTAL_MINUTES = divmod(c.days * 86400 + c.seconds, 60)[0]

# At 1 block per minute
TOTAL_BLOCKS = Decimal(TOTAL_MINUTES)


def calc_coeff() -> Decimal:
    """
    block reward calculation. Decay curve: 200 years
    >>> calc_coeff()
    Decimal('1.664087503734056374552843909E-7')
    """
    return config.dev.coin_remaning_at_genesis.ln() / TOTAL_BLOCKS


def remaining_emission(block_n) -> Decimal:
    # TODO: This is more related to the way QRL works.. Move to another place
    """
    calculate remaining emission at block_n: N=total initial coin supply, coeff = decay constant
    need to use decimal as floating point not precise enough on different platforms..
    :param block_n:
    :return:

    >>> remaining_emission(0)
    Decimal('40000000000000000')
    >>> remaining_emission(1)
    Decimal('39999993343650538')
    >>> remaining_emission(2)
    Decimal('39999986687302185')
    >>> remaining_emission(100)
    Decimal('39999334370536850')
    """
    coeff = calc_coeff()
    return (config.dev.coin_remaning_at_genesis * config.dev.shor_per_quanta * Decimal(-coeff * block_n).exp()) \
        .quantize(Decimal('1.'), rounding=decimal.ROUND_DOWN)


def block_reward(block_number) -> Decimal:
    """
    :return: Block reward in shors for block number

    >>> block_reward(1)
    Decimal('6656349462')
    >>> block_reward(2)
    Decimal('6656348353')
    >>> block_reward(3)
    Decimal('6656347246')
    >>> N = 40
    >>> tmp_sum = sum(block_reward(b) for b in range(1, N))
    >>> tmp_est = remaining_emission(0) - remaining_emission(N-1)
    >>> tmp_est - tmp_sum
    Decimal('0')
    """
    return remaining_emission(block_number - 1) - remaining_emission(block_number)



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
import random

from pyqrllib.pyqrllib import bin2hstr
from pyqryptonight.pyqryptonight import UInt256ToString
from twisted.internet import reactor
from twisted.internet.protocol import ServerFactory

from qrl.core import config
from qrl.core.Block import Block
from qrl.core.ChainManager import ChainManager
from qrl.core.ESyncState import ESyncState
from qrl.core.messagereceipt import MessageReceipt
from qrl.core.misc import ntp, logger
from qrl.core.node import SyncState
from qrl.core.p2p.p2pprotocol import P2PProtocol
from qrl.core.p2p.IPMetadata import IPMetadata
from qrl.core.processors.TxnProcessor import TxnProcessor
from qrl.core.txs.MessageTransaction import MessageTransaction
from qrl.core.txs.SlaveTransaction import SlaveTransaction
from qrl.core.txs.TokenTransaction import TokenTransaction
from qrl.core.txs.TransferTokenTransaction import TransferTokenTransaction
from qrl.core.txs.TransferTransaction import TransferTransaction
from qrl.generated import qrllegacy_pb2, qrl_pb2

p2p_msg_priority = {
            qrllegacy_pb2.LegacyMessage.VE: 0,
            qrllegacy_pb2.LegacyMessage.PL: 0,
            qrllegacy_pb2.LegacyMessage.PONG: 0,

            ######################
            qrllegacy_pb2.LegacyMessage.MR: 2,
            qrllegacy_pb2.LegacyMessage.SFM: 1,

            qrllegacy_pb2.LegacyMessage.BK: 1,
            qrllegacy_pb2.LegacyMessage.FB: 0,
            qrllegacy_pb2.LegacyMessage.PB: 0,
            qrllegacy_pb2.LegacyMessage.BH: 1,

            ############################
            qrllegacy_pb2.LegacyMessage.TX: 1,
            qrllegacy_pb2.LegacyMessage.MT: 1,
            qrllegacy_pb2.LegacyMessage.TK: 1,
            qrllegacy_pb2.LegacyMessage.TT: 1,
            qrllegacy_pb2.LegacyMessage.LT: 1,
            qrllegacy_pb2.LegacyMessage.SL: 1,

            qrllegacy_pb2.LegacyMessage.EPH: 3,

            qrllegacy_pb2.LegacyMessage.SYNC: 0,
            qrllegacy_pb2.LegacyMessage.CHAINSTATE: 0,
            qrllegacy_pb2.LegacyMessage.HEADERHASHES: 1,
            qrllegacy_pb2.LegacyMessage.P2P_ACK: 0,
        }


class P2PFactory(ServerFactory):
    protocol = P2PProtocol

    def __init__(self,
                 chain_manager: ChainManager,
                 sync_state: SyncState,
                 qrl_node):

        self.master_mr = MessageReceipt()
        self.pow = None
        self.sync_state = sync_state

        self._ntp = ntp
        self._qrl_node = qrl_node
        self._chain_manager = chain_manager
        self._chain_manager.set_broadcast_tx(self.broadcast_tx)

        self._syncing_enabled = False
        self._target_channel = None
        self._target_node_header_hash = None
        self._last_requested_block_number = None

        self._genesis_processed = False
        self._peer_connections = []
        self._txn_processor_running = False

        self.peer_blockheight = dict()

        reactor.callLater(config.user.monitor_connections_interval, self.monitor_connections)

        self.p2p_msg_priority = p2p_msg_priority

    ###################################################
    ###################################################
    ###################################################
    ###################################################
    ###################################################
    ###################################################

    def get_random_peer(self):
        # FIXME: Used a named tuple to improve readability?
        # FIXME: This probably can go the peerManager
        max_cumulative_difficulty = 0
        for addr_remote in self.peer_blockheight:
            max_cumulative_difficulty = max(max_cumulative_difficulty, self.peer_blockheight[addr_remote][2])

        best_connection_ids = []
        for addr_remote in self.peer_blockheight:
            if self.peer_blockheight[addr_remote][2] == max_cumulative_difficulty:
                best_connection_ids.append(addr_remote)

        selected_peer_connections = []
        for addr_remote in best_connection_ids:
            for peer_conn in self._peer_connections:
                if peer_conn.peer.full_address == addr_remote:
                    selected_peer_connections.append(peer_conn)

        if len(selected_peer_connections) == 0 or max_cumulative_difficulty == 0:
            return None

        return random.sample(selected_peer_connections, 1)[0]

    def update_peer_blockheight(self, addr_remote, block_number, headerhash, cumulative_difficulty):
        # FIXME: Use a named tuple to improve readability?
        self.peer_blockheight[addr_remote] = [block_number, headerhash, int(UInt256ToString(cumulative_difficulty))]

    def request_peer_blockheight(self):
        for peer in self._peer_connections:
            msg = qrllegacy_pb2.LegacyMessage(func_name=qrllegacy_pb2.LegacyMessage.BH,
                                              bhData=qrl_pb2.BlockHeightData(block_number=0))
            peer.send(msg)

    ###################################################
    ###################################################
    ###################################################
    ###################################################
    ###################################################
    ###################################################
    ###################################################

    @property
    def num_connections(self):
        return len(self._peer_connections)

    @property
    def connections(self):
        return list(self._peer_connections)

    @property
    def synced(self):
        return self.pow.sync_state.state == ESyncState.synced

    @property
    def reached_conn_limit(self):
        return len(self._peer_connections) >= config.user.max_peers_limit

    def get_connected_peer_addrs(self):
        return set([peer.peer.full_address for peer in self._peer_connections])

    ###################################################
    ###################################################
    ###################################################
    ###################################################
    ###################################################
    ###################################################

    @property
    def chain_height(self):
        return self._chain_manager.height

    def get_last_block(self):
        return self._chain_manager.last_block

    def get_headerhashes(self, start_blocknumber):
        return self._chain_manager.get_headerhashes(start_blocknumber)

    def get_cumulative_difficulty(self):
        return self._chain_manager.get_cumulative_difficulty()

    def get_block_by_number(self, block_number):
        return self._chain_manager.get_block_by_number(block_number)

    def is_block_present(self, header_hash: bytes) -> bool:
        if not self._chain_manager.get_block(header_hash):
            if header_hash not in self.pow.future_blocks:
                return False

        return True

    def block_received(self, source, block: Block):
        self.pow.last_pb_time = ntp.getTime()
        logger.info('>>> Received Block #%d %s', block.block_number, bin2hstr(block.headerhash))

        if source != self._target_channel:
            if self._target_channel is None:
                logger.warning('Received block and target channel is None')
            else:
                logger.warning('Received block from unexpected peer')
                logger.warning('Expected peer: %s', self._target_channel.peer)
                logger.warning('Found peer: %s', source.peer)
            return

        if block.block_number != self._last_requested_block_number:
            logger.warning('Did not match %s', self._last_requested_block_number)
            return

        target_start_blocknumber = self._target_node_header_hash.block_number
        expected_headerhash = self._target_node_header_hash.headerhashes[block.block_number - target_start_blocknumber]
        if block.headerhash != expected_headerhash:
            logger.warning('Did not match headerhash')
            logger.warning('Expected headerhash %s', expected_headerhash)
            logger.warning('Found headerhash %s', block.headerhash)
            return

        if not block.validate(self._chain_manager, self.pow.future_blocks):
            logger.warning('Syncing Failed: Block Validation Failed')
            return

        if self._chain_manager.add_block(block, check_stale=False):
            if self._chain_manager.last_block.headerhash == block.headerhash:
                self.pow.suspend_mining_timestamp = ntp.getTime() + config.dev.sync_delay_mining
        else:
            logger.warning('Failed to Add Block')
            return

        try:
            reactor.download_monitor.cancel()
        except Exception as e:
            logger.warning("PB: %s", e)

        if self.is_syncing_finished():
            return

        self._last_requested_block_number += 1

        self.peer_fetch_block()

    def is_syncing(self) -> bool:
        return self._syncing_enabled

    def is_syncing_finished(self, force_finish=False):
        curr_index = self._last_requested_block_number - self._target_node_header_hash.block_number + 1
        if curr_index == len(self._target_node_header_hash.headerhashes) or force_finish:
            self._last_requested_block_number = None
            self._target_node_header_hash = None
            self._target_channel = None
            self._syncing_enabled = False
            return True

        return False

    def peer_fetch_block(self, retry=0):
        node_header_hash = self._target_node_header_hash
        curr_index = self._last_requested_block_number - node_header_hash.block_number

        block_headerhash = node_header_hash.headerhashes[curr_index]
        block = self._chain_manager.get_block(block_headerhash)

        if retry >= 1:
            logger.debug('Retry Limit Hit')
            self._qrl_node.peer_manager.ban_channel(self._target_channel)
            self.is_syncing_finished(force_finish=True)
            return

        while block and curr_index + 1 < len(node_header_hash.headerhashes):
            self._last_requested_block_number += 1
            curr_index = self._last_requested_block_number - node_header_hash.block_number
            block_headerhash = node_header_hash.headerhashes[curr_index]
            block = self._chain_manager.get_block(block_headerhash)

        if block and self.is_syncing_finished():
            return

        self._target_channel.send_fetch_block(self._last_requested_block_number)
        reactor.download_monitor = reactor.callLater(100, self.peer_fetch_block, retry + 1)

    def compare_and_sync(self, source_peer, node_header_hash: qrl_pb2.NodeHeaderHash):
        if self._syncing_enabled:
            logger.info('>> Ignoring compare_and_sync Syncing Enabled')
            return
        last_block = self.get_last_block()
        node_last_block_number = node_header_hash.block_number + len(node_header_hash.headerhashes) - 1
        last_block_number = min(last_block.block_number, node_last_block_number)
        if last_block_number < node_header_hash.block_number:
            return
        fork_block_number = last_block.block_number + 1
        fork_found = False
        for i in range(last_block_number, node_header_hash.block_number - 1, -1):
            block = self._chain_manager.get_block_by_number(i)
            if block:
                if block.headerhash == node_header_hash.headerhashes[i - node_header_hash.block_number]:
                    break
            fork_block_number = i
            fork_found = True

        if fork_found or (last_block.block_number < node_last_block_number):
            self._target_channel = source_peer
            self._target_node_header_hash = node_header_hash
            self._last_requested_block_number = fork_block_number
            self._syncing_enabled = True
            self.peer_fetch_block()

    ###################################################
    ###################################################
    ###################################################
    ###################################################
    ###################################################
    ###################################################

    def request_full_message(self, mr_data: qrllegacy_pb2.MRData):
        """
        Request Full Message
        This function request for the full message against,
        the Message Receipt received.
        :return:
        """

        # FIXME: Again, breaking encasulation
        # FIXME: Huge amount of lookups in dictionaries
        msg_hash = mr_data.hash

        if msg_hash in self.master_mr._hash_msg:
            if msg_hash in self.master_mr.requested_hash:
                del self.master_mr.requested_hash[msg_hash]
            return

        if msg_hash not in self.master_mr.requested_hash:
            return

        peers_list = self.master_mr.requested_hash[msg_hash].peers_connection_list
        message_request = self.master_mr.requested_hash[msg_hash]
        for peer in peers_list:
            if peer in message_request.already_requested_peers:
                continue
            message_request.already_requested_peers.append(peer)

            msg = qrllegacy_pb2.LegacyMessage(func_name=qrllegacy_pb2.LegacyMessage.SFM,
                                              mrData=qrllegacy_pb2.MRData(hash=mr_data.hash, type=mr_data.type))

            peer.send(msg)

            call_later_obj = reactor.callLater(config.dev.message_receipt_timeout,
                                               self.request_full_message,
                                               mr_data)

            message_request.callLater = call_later_obj
            return

        # If execution reach to this line, then it means no peer was able to provide
        # Full message for this hash thus the hash has to be deleted.
        # Moreover, negative points could be added to the peers, for this behavior
        if msg_hash in self.master_mr.requested_hash:
            del self.master_mr.requested_hash[msg_hash]

    ##############################################
    ##############################################
    ##############################################
    ##############################################

    def reset_processor_flag(self, _):
        self._txn_processor_running = False

    def reset_processor_flag_with_err(self, msg):
        logger.error('Exception in txn task')
        logger.error('%s', msg)
        self._txn_processor_running = False

    def add_unprocessed_txn(self, tx, ip) -> bool:
        if tx.fee < config.user.transaction_minimum_fee:
            logger.info("Dropping Txn %s", bin2hstr(tx.txhash))
            logger.info("Reason: Fee %s is below threshold fee %s", tx.fee, config.user.transaction_minimum_fee)
            return False

        if not self._chain_manager.tx_pool.update_pending_tx_pool(tx, ip):
            return False

        if not self._txn_processor_running:
            txn_processor = TxnProcessor(chain_manager=self._chain_manager,
                                         transaction_pool_obj=self._chain_manager.tx_pool,
                                         broadcast_tx=self.broadcast_tx)

            task_defer = TxnProcessor.create_cooperate(txn_processor).whenDone()
            task_defer.addCallback(self.reset_processor_flag) \
                .addErrback(self.reset_processor_flag_with_err)
            self._txn_processor_running = True

        return True

    ##############################################
    ##############################################
    ##############################################
    ##############################################

    def broadcast_tx(self, tx: TransferTransaction):
        logger.info('<<<Transmitting TX: %s', bin2hstr(tx.txhash))

        if isinstance(tx, MessageTransaction):
            legacy_type = qrllegacy_pb2.LegacyMessage.MT
        elif isinstance(tx, TransferTransaction):
            legacy_type = qrllegacy_pb2.LegacyMessage.TX
        elif isinstance(tx, TokenTransaction):
            legacy_type = qrllegacy_pb2.LegacyMessage.TK
        elif isinstance(tx, TransferTokenTransaction):
            legacy_type = qrllegacy_pb2.LegacyMessage.TT
        elif isinstance(tx, SlaveTransaction):
            legacy_type = qrllegacy_pb2.LegacyMessage.SL
        else:
            raise ValueError('Invalid Transaction Type')
        self.register_and_broadcast(legacy_type, tx.get_message_hash(), tx.pbdata)

    def broadcast_block(self, block: Block):
        # logger.info('<<<Transmitting block: ', block.headerhash)
        data = qrllegacy_pb2.MRData()
        data.stake_selector = block.transactions[0].public_key
        data.block_number = block.block_number
        data.prev_headerhash = bytes(block.prev_headerhash)

        self.register_and_broadcast(qrllegacy_pb2.LegacyMessage.BK, block.headerhash, block.pbdata, data)

    def register_and_broadcast(self, msg_type, msg_hash: bytes, pbdata, data=None):
        self.master_mr.register(msg_type, msg_hash, pbdata)
        self.broadcast(msg_type, msg_hash, data)

    def broadcast(self, msg_type, msg_hash: bytes, mr_data=None):
        """
        Broadcast
        This function sends the Message Receipt to all connected peers.
        :return:
        """
        ignore_peers = []
        if msg_hash in self.master_mr.requested_hash:
            ignore_peers = self.master_mr.requested_hash[msg_hash].peers_connection_list

        if not mr_data:
            mr_data = qrllegacy_pb2.MRData()

        mr_data.hash = msg_hash
        mr_data.type = msg_type
        data = qrllegacy_pb2.LegacyMessage(func_name=qrllegacy_pb2.LegacyMessage.MR,
                                           mrData=mr_data)

        for peer in self._peer_connections:
            if peer not in ignore_peers:
                peer.send(data)

    def broadcast_get_synced_state(self):
        # Request all peers to update their synced status
        for peer in self._peer_connections:
            peer.send_sync()

    ###################################################
    ###################################################
    ###################################################
    ###################################################
    # Event handlers / Comms related

    def start_listening(self):
        reactor.listenTCP(config.user.p2p_local_port, self)

    def clientConnectionLost(self, connector, reason):  # noqa
        logger.debug('connection lost: %s', reason)

    def clientConnectionFailed(self, connector, reason):
        logger.debug('connection failed: %s', reason)

    def startedConnecting(self, connector):
        logger.debug('Started connecting: %s', connector)

    def add_connection(self, conn_protocol) -> bool:
        # TODO: Most of this can go peer manager
        if self._qrl_node.peer_manager.is_banned(conn_protocol.peer):
            return False

        redundancy_count = 0
        for conn in self._peer_connections:
            if conn.peer.ip == conn_protocol.peer.ip:
                redundancy_count += 1

        if config.user.max_redundant_connections >= 0:
            if redundancy_count >= config.user.max_redundant_connections:
                logger.info('Redundant Limit. Disconnecting client %s', conn_protocol.peer)
                return False

        if self.reached_conn_limit:
            # FIXME: Should we stop listening to avoid unnecessary load due to many connections?
            logger.info('Peer limit hit. Disconnecting client %s', conn_protocol.peer)
            return False

        if conn_protocol.peer.ip == conn_protocol.host.ip and conn_protocol.peer.port == config.user.p2p_public_port:
            peer_list = [p for p in self._qrl_node.peer_manager.known_peer_addresses if p != conn_protocol.peer.full_address]
            self._qrl_node.peer_manager.extend_known_peers(peer_list)
            return False

        self._peer_connections.append(conn_protocol)

        logger.debug('>>> new connection: %s ', conn_protocol.peer)
        return True

    def remove_connection(self, conn_protocol):
        if conn_protocol in self._peer_connections:
            self._peer_connections.remove(conn_protocol)

        if conn_protocol.peer.full_address in self.peer_blockheight:
            del self.peer_blockheight[conn_protocol.peer.full_address]

    def monitor_connections(self):
        reactor.callLater(config.user.monitor_connections_interval, self.monitor_connections)

        if len(self._peer_connections) == 0:
            logger.warning('No Connected Peer Found')
            reactor.callLater(10, self._qrl_node.peer_manager.connect_peers)
            return

        connected_peers_set = set()
        for conn_protocol in self._peer_connections:
            connected_peers_set.add(conn_protocol.peer.full_address)

        for peer_item in config.user.peer_list:
            peer_metadata = IPMetadata.from_full_address(peer_item)
            if peer_metadata.full_address not in connected_peers_set:
                self.connect_peer(peer_metadata.full_address)

    def connect_peer(self, full_address):
        try:
            addr = IPMetadata.from_full_address(full_address)

            connected_peers = self.get_connected_peer_addrs()
            should_connect = addr.full_address not in connected_peers

            if should_connect:
                reactor.connectTCP(addr.ip, addr.port, self)

        except Exception as e:
            logger.warning("Could not connect to %s - %s", full_address, str(e))



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
import os

import yaml
import simplejson as json
from google.protobuf.json_format import Parse

from qrl.core import config
from qrl.core.Block import Block
from qrl.core.Singleton import Singleton
from qrl.generated import qrl_pb2


class GenesisBlock(Block, metaclass=Singleton):
    def __init__(self):
        package_directory = os.path.dirname(os.path.abspath(__file__))

        genesis_data_path = os.path.join(package_directory, 'genesis.yml')
        genesis_config_path = os.path.join(config.user.qrl_dir, 'genesis.yml')

        if os.path.isfile(genesis_config_path):
            with open(genesis_config_path) as f:
                genesisBlock_json = json.dumps(yaml.safe_load(f))
        else:
            with open(genesis_data_path) as f:
                genesisBlock_json = json.dumps(yaml.safe_load(f))

        tmp_block = qrl_pb2.Block()
        Parse(genesisBlock_json, tmp_block)
        super(GenesisBlock, self).__init__(tmp_block)

    @property
    def transactions(self):
        return self._data.transactions

    @property
    def genesis_balance(self):
        return self._data.genesis_balance



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
import os
from enum import Enum
from typing import Callable, Set, List

import simplejson as json
from pyqryptonight.pyqryptonight import UInt256ToString

from qrl.core import config
from qrl.core.misc import logger, ntp
from qrl.core.misc.expiring_set import ExpiringSet
from qrl.core.notification.Observable import Observable
from qrl.core.notification.ObservableEvent import ObservableEvent
from qrl.core.p2p.IPMetadata import IPMetadata
from qrl.core.p2p.p2pObserver import P2PBaseObserver
from qrl.core.p2p.p2pprotocol import P2PProtocol
from qrl.generated import qrllegacy_pb2, qrl_pb2


class P2PPeerManager(P2PBaseObserver):
    class EventType(Enum):
        NO_PEERS = 1

    def __init__(self):
        super().__init__()
        self._ping_callLater = None
        self._disconnect_callLater = None
        self._channels = []

        self._peer_node_status = dict()

        self._known_peers = set()
        self.peers_path = os.path.join(config.user.data_dir,
                                       config.dev.peers_filename)

        self.banned_peers_filename = os.path.join(config.user.wallet_dir, config.dev.banned_peers_filename)
        self._banned_peer_ips = ExpiringSet(expiration_time=config.user.ban_minutes * 60,
                                            filename=self.banned_peers_filename)

        self._observable = Observable(self)
        self._p2pfactory = None

    def register(self, message_type: EventType, func: Callable):
        self._observable.register(message_type, func)

    @property
    def known_peer_addresses(self):
        return self._known_peers

    def trusted_peer(self, channel: P2PProtocol):
        if self.is_banned(channel.peer):
            return False

        if channel.valid_message_count < config.dev.trust_min_msgcount:
            return False

        if channel.connection_time < config.dev.trust_min_conntime:
            return False

        return True

    @property
    def trusted_addresses(self):
        return set([peer.peer.full_address for peer in self._p2pfactory.connections if self.trusted_peer(peer)])

    @property
    def peer_node_status(self):
        return self._peer_node_status

    def load_known_peers(self) -> List[str]:
        known_peers = []
        try:
            logger.info('Loading known peers')
            with open(self.peers_path, 'r') as infile:
                known_peers = json.load(infile)
        except Exception as e:
            logger.info("Could not open known_peers list")

        return [IPMetadata.canonical_full_address(fa) for fa in known_peers]

    def save_known_peers(self, known_peers: List[str]):
        tmp = list(known_peers)
        config.create_path(config.user.data_dir)
        with open(self.peers_path, 'w') as outfile:
            json.dump(tmp, outfile)

    def load_peer_addresses(self) -> None:
        known_peers = self.load_known_peers()
        self._known_peers = self.combine_peer_lists(known_peers, config.user.peer_list, )
        logger.info('Loaded known peers: %s', self._known_peers)
        self.save_known_peers(self._known_peers)

    def extend_known_peers(self, new_peer_addresses: set) -> None:
        new_addresses = set(new_peer_addresses) - self._known_peers

        if self._p2pfactory is not None:
            for peer_address in new_addresses:
                self._p2pfactory.connect_peer(peer_address)

        self._known_peers |= set(new_peer_addresses)
        self.save_known_peers(list(self._known_peers))

    @staticmethod
    def combine_peer_lists(peer_ips, sender_full_addresses: List, check_global=False) -> Set[IPMetadata]:
        tmp_list = list(peer_ips)
        tmp_list.extend(sender_full_addresses)

        answer = set()
        for item in tmp_list:
            try:
                answer.add(IPMetadata.canonical_full_address(item, check_global))
            except:  # noqa
                logger.warning("Invalid Peer Address {}".format(item))

        return answer

    def get_better_difficulty(self, current_cumulative_difficulty):
        best_cumulative_difficulty = int(UInt256ToString(current_cumulative_difficulty))
        local_best = best_cumulative_difficulty
        best_channel = None

        for channel in self._peer_node_status:
            node_chain_state = self._peer_node_status[channel]
            node_cumulative_difficulty = int(UInt256ToString(node_chain_state.cumulative_difficulty))
            if node_cumulative_difficulty > best_cumulative_difficulty:
                best_cumulative_difficulty = node_cumulative_difficulty
                best_channel = channel
        logger.debug('Local Best Diff : %s', local_best)
        logger.debug('Remote Best Diff : %s', best_cumulative_difficulty)
        return best_channel

    def remove_channel(self, channel):
        if channel in self._channels:
            self._channels.remove(channel)
        if channel in self._peer_node_status:
            del self._peer_node_status[channel]

    def new_channel(self, channel):
        self._channels.append(channel)
        self._peer_node_status[channel] = qrl_pb2.NodeChainState(block_number=0,
                                                                 header_hash=b'',
                                                                 cumulative_difficulty=b'\x00' * 32,
                                                                 timestamp=ntp.getTime())
        channel.register(qrllegacy_pb2.LegacyMessage.VE, self.handle_version)
        channel.register(qrllegacy_pb2.LegacyMessage.PL, self.handle_peer_list)
        channel.register(qrllegacy_pb2.LegacyMessage.CHAINSTATE, self.handle_chain_state)
        channel.register(qrllegacy_pb2.LegacyMessage.SYNC, self.handle_sync)
        channel.register(qrllegacy_pb2.LegacyMessage.P2P_ACK, self.handle_p2p_acknowledgement)

    def handle_version(self, source, message: qrllegacy_pb2.LegacyMessage):
        """
        Version
        If version is empty, it sends the version & genesis_prev_headerhash.
        Otherwise, processes the content of data.
        In case of mismatches, it disconnects from the peer
        """
        self._validate_message(message, qrllegacy_pb2.LegacyMessage.VE)

        if not message.veData.version:
            msg = qrllegacy_pb2.LegacyMessage(
                func_name=qrllegacy_pb2.LegacyMessage.VE,
                veData=qrllegacy_pb2.VEData(version=config.dev.version,
                                            genesis_prev_hash=config.user.genesis_prev_headerhash,
                                            rate_limit=config.user.peer_rate_limit))

            source.send(msg)
            return

        logger.info('%s version: %s | genesis prev_headerhash %s',
                    source.peer.ip,
                    message.veData.version,
                    message.veData.genesis_prev_hash)

        source.rate_limit = min(config.user.peer_rate_limit, message.veData.rate_limit)

        if message.veData.genesis_prev_hash != config.user.genesis_prev_headerhash:
            logger.warning('%s genesis_prev_headerhash mismatch', source.peer)
            logger.warning('Expected: %s', config.user.genesis_prev_headerhash)
            logger.warning('Found: %s', message.veData.genesis_prev_hash)
            source.loseConnection()

    def handle_peer_list(self, source, message: qrllegacy_pb2.LegacyMessage):
        P2PBaseObserver._validate_message(message, qrllegacy_pb2.LegacyMessage.PL)

        if not config.user.enable_peer_discovery:
            return

        if not message.plData.peer_ips:
            return

        sender_peer = IPMetadata(source.peer.ip, message.plData.public_port)

        new_peers = self.combine_peer_lists(message.plData.peer_ips, [sender_peer.full_address], check_global=True)
        new_peers.discard(source.host.full_address)  # Remove local address

        logger.info('%s peers data received: %s', source.peer.ip, new_peers)
        self.extend_known_peers(new_peers)

    def handle_sync(self, source, message: qrllegacy_pb2.LegacyMessage):
        P2PBaseObserver._validate_message(message, qrllegacy_pb2.LegacyMessage.SYNC)
        if message.syncData.state == '':
            if source.factory.synced:
                source.send_sync(synced=True)

    @staticmethod
    def send_node_chain_state(dest_channel, node_chain_state: qrl_pb2.NodeChainState):
        # FIXME: Not sure this belongs to peer management
        msg = qrllegacy_pb2.LegacyMessage(func_name=qrllegacy_pb2.LegacyMessage.CHAINSTATE,
                                          chainStateData=node_chain_state)
        dest_channel.send(msg)

    def monitor_chain_state(self):
        # FIXME: Not sure this belongs to peer management
        current_timestamp = ntp.getTime()
        for channel in self._channels:
            if channel not in self._peer_node_status:
                channel.loseConnection()
                continue
            delta = current_timestamp - self._peer_node_status[channel].timestamp
            if delta > config.user.chain_state_timeout:
                del self._peer_node_status[channel]
                logger.debug('>>>> No State Update [%18s] %2.2f (TIMEOUT)', channel.peer, delta)
                channel.loseConnection()

    def broadcast_chain_state(self, node_chain_state: qrl_pb2.NodeChainState):
        # FIXME: Not sure this belongs to peer management
        # TODO: Verify/Disconnect problematic channels
        # Ping all channels
        for channel in self._channels:
            self.send_node_chain_state(channel, node_chain_state)

        self._observable.notify(ObservableEvent(self.EventType.NO_PEERS))

    def handle_chain_state(self, source, message: qrllegacy_pb2.LegacyMessage):
        # FIXME: Not sure this belongs to peer management
        P2PBaseObserver._validate_message(message, qrllegacy_pb2.LegacyMessage.CHAINSTATE)

        message.chainStateData.timestamp = ntp.getTime()  # Receiving time

        try:
            UInt256ToString(message.chainStateData.cumulative_difficulty)
        except ValueError:
            logger.warning('Invalid Cumulative Difficulty sent by peer')
            source.loseConnection()
            return

        self._peer_node_status[source] = message.chainStateData

    def handle_p2p_acknowledgement(self, source, message: qrllegacy_pb2.LegacyMessage):
        P2PBaseObserver._validate_message(message, qrllegacy_pb2.LegacyMessage.P2P_ACK)

        source.bytes_sent -= message.p2pAckData.bytes_processed
        if source.bytes_sent < 0:
            logger.warning('Disconnecting Peer %s', source.peer)
            logger.warning('Reason: negative bytes_sent value')
            logger.warning('bytes_sent %s', source.bytes_sent)
            logger.warning('Ack bytes_processed %s', message.p2pAckData.bytes_processed)
            source.loseConnection()

        source.send_next()

    ####################################################
    ####################################################
    ####################################################
    ####################################################
    def is_banned(self, peer: IPMetadata):
        return peer.ip in self._banned_peer_ips

    def ban_channel(self, channel: P2PProtocol):
        self._banned_peer_ips.add(channel.peer.ip)
        logger.warning('Banned %s', channel.peer.ip)
        channel.loseConnection()

    def connect_peers(self):
        logger.info('<<<Reconnecting to peer list: %s', self.known_peer_addresses)
        for peer_address in self.known_peer_addresses:
            if self.is_banned(IPMetadata.from_full_address(peer_address)):
                continue
            self._p2pfactory.connect_peer(peer_address)

    def get_peers_stat(self) -> list:
        peers_stat = []
        # Copying the list of keys, to avoid any change by other thread
        for source in list(self.peer_node_status.keys()):
            try:
                peer_stat = qrl_pb2.PeerStat(peer_ip=source.peer.ip.encode(),
                                             port=source.peer.port,
                                             node_chain_state=self.peer_node_status[source])
                peers_stat.append(peer_stat)
            except KeyError:
                # Ignore in case the key is deleted by other thread causing KeyError
                continue
        return peers_stat



from importlib import import_module

TYPENAME_MAP = {
    'transfer': 'TransferTransaction',
    'coinbase': 'CoinBase',
    'latticePK': 'LatticePublicKey',
    'message': 'MessageTransaction',
    'token': 'TokenTransaction',
    'transfer_token': 'TransferTokenTransaction',
    'slave': 'SlaveTransaction'
}


def build_tx(pb_tx_type, *args, **kwargs):
    try:
        tx_class_name = TYPENAME_MAP[pb_tx_type]
        tx_module = import_module('.' + tx_class_name, package='qrl.core.txs')
        tx_class = getattr(tx_module, tx_class_name)
        return tx_class(*args, **kwargs)

    except(AttributeError, ModuleNotFoundError) as e:  # noqa
        raise ImportError("{} is not defined as a transaction type".format(pb_tx_type))



from unittest import TestCase

import simplejson as json
from mock import patch, Mock, PropertyMock
from pyqrllib.pyqrllib import bin2hstr

from qrl.core import config
from qrl.core.misc import logger
from qrl.core.AddressState import AddressState
from qrl.core.BlockHeader import BlockHeader
from qrl.core.ChainManager import ChainManager
from qrl.core.txs.CoinBase import CoinBase
from qrl.crypto.misc import sha256
from tests.core.txs.testdata import test_json_CoinBase
from tests.misc.helper import get_alice_xmss

logger.initialize_default()


@patch('qrl.core.txs.Transaction.logger')
class TestCoinBase(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestCoinBase, self).__init__(*args, **kwargs)

        self.alice = get_alice_xmss()
        self.alice.set_ots_index(11)

        self.mock_blockheader = Mock(spec=BlockHeader)
        self.mock_blockheader.stake_selector = self.alice.address
        self.mock_blockheader.block_reward = 50
        self.mock_blockheader.fee_reward = 40
        self.mock_blockheader.prev_headerhash = sha256(b'prev_headerhash')
        self.mock_blockheader.block_number = 1
        self.mock_blockheader.headerhash = sha256(b'headerhash')

        self.amount = self.mock_blockheader.block_reward + self.mock_blockheader.fee_reward
        self.maxDiff = None

    def test_create(self, m_logger):
        tx = CoinBase.create(self.amount, self.alice.address, self.mock_blockheader.block_number)
        self.assertIsInstance(tx, CoinBase)

    def test_to_json(self, m_logger):
        amount = self.mock_blockheader.block_reward + self.mock_blockheader.fee_reward
        tx = CoinBase.create(amount, self.alice.address, self.mock_blockheader.block_number)
        txjson = tx.to_json()
        self.assertEqual(json.loads(test_json_CoinBase), json.loads(txjson))

    def test_from_txdict(self, m_logger):
        amount = self.mock_blockheader.block_reward + self.mock_blockheader.fee_reward
        tx = CoinBase.create(amount, self.alice.address, self.mock_blockheader.block_number)
        self.assertIsInstance(tx, CoinBase)

        # Test that common Transaction components were copied over.
        self.assertEqual(self.mock_blockheader.block_number + 1, tx.nonce)
        self.assertEqual('010300a1da274e68c88b0ccf448e0b1916fa789b01eb2ed4e9ad565ce264c9390782a9c61ac02f',
                         bin2hstr(tx.addr_to))

        self.assertEqual('222460cc57ab8683b46f1831fe6cf1832c7e3134baf74d33bfaf91741e19cba2', bin2hstr(tx.txhash))
        self.assertEqual(tx.amount, 90)

    def test_validate_custom(self, m_logger):
        """
        CoinBase _validate_custom() only checks if fee == 0
        """
        tx = CoinBase.create(self.amount, self.alice.address, self.mock_blockheader.block_number)
        tx._data.fee = 1
        result = tx._validate_custom()
        self.assertFalse(result)

        tx._data.fee = 0
        result = tx._validate_custom()
        self.assertTrue(result)

    def test_validate_extended(self, m_logger):
        """
        CoinBase validate_extended() checks for
        1. valid coinbase address (the coinbase address must be config.dev.coinbase_address)
        2. valid addr_to
        then calls _validate_custom()
        """
        tx = CoinBase.create(self.amount, self.alice.address, self.mock_blockheader.block_number)
        tx._data.master_addr = self.alice.address

        result = tx.validate_extended(self.mock_blockheader.block_number)
        self.assertFalse(result)

        tx._data.master_addr = config.dev.coinbase_address
        with patch('qrl.core.txs.CoinBase.CoinBase.addr_to', new_callable=PropertyMock) as m_addr_to:
            m_addr_to.return_value = b'Fake Address'
            result = tx.validate_extended(self.mock_blockheader.block_number)
            self.assertFalse(result)

        result = tx.validate_extended(self.mock_blockheader.block_number)
        self.assertTrue(result)

    def test_apply_state_changes(self, m_logger):
        """
        Alice earned some coins.
        """
        addresses_state = {
            config.dev.coinbase_address: Mock(autospec=AddressState, name='CoinBase AddressState',
                                              transaction_hashes=[], balance=1000000),
            self.alice.address: Mock(autospec=AddressState, name='alice AddressState', transaction_hashes=[],
                                     balance=0),
        }
        tx = CoinBase.create(self.amount, self.alice.address, self.mock_blockheader.block_number)

        tx.apply_state_changes(addresses_state)

        self.assertEqual(1000000 - tx.amount, addresses_state[config.dev.coinbase_address].balance)
        self.assertEqual([tx.txhash], addresses_state[config.dev.coinbase_address].transaction_hashes)
        self.assertEqual(tx.amount, addresses_state[self.alice.address].balance)
        self.assertEqual([tx.txhash], addresses_state[self.alice.address].transaction_hashes)

        # A blank addresses_state doesn't get modified at all (but in practice, every node should have an AddressState
        # for the CoinBase addr
        addresses_state_empty = {}
        tx.apply_state_changes(addresses_state_empty)
        self.assertEqual({}, addresses_state_empty)

    def test_revert_state_changes(self, m_logger):
        """
        Alice earned some coins. Undo this.
        """
        tx = CoinBase.create(self.amount, self.alice.address, self.mock_blockheader.block_number)
        addresses_state = {
            config.dev.coinbase_address: Mock(autospec=AddressState, name='CoinBase AddressState',
                                              transaction_hashes=[tx.txhash], balance=1000000 - self.amount),
            self.alice.address: Mock(autospec=AddressState, name='alice AddressState', transaction_hashes=[tx.txhash],
                                     balance=self.amount),
        }
        unused_chain_manager_mock = Mock(autospec=ChainManager, name='unused ChainManager')

        tx.revert_state_changes(addresses_state, unused_chain_manager_mock)

        self.assertEqual(1000000, addresses_state[config.dev.coinbase_address].balance)
        self.assertEqual([], addresses_state[config.dev.coinbase_address].transaction_hashes)
        self.assertEqual(0, addresses_state[self.alice.address].balance)
        self.assertEqual([], addresses_state[self.alice.address].transaction_hashes)

        # A blank addresses_state doesn't get modified at all (but in practice, every node should have an AddressState
        # for the CoinBase addr
        addresses_state_empty = {}
        tx.revert_state_changes(addresses_state_empty, unused_chain_manager_mock)
        self.assertEqual({}, addresses_state_empty)

    def test_affected_address(self, m_logger):
        # This transaction can only involve 2 addresses.
        affected_addresses = set()
        tx = CoinBase.create(self.amount, self.alice.address, self.mock_blockheader.block_number)
        tx.set_affected_address(affected_addresses)
        self.assertEqual(2, len(affected_addresses))



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from unittest import TestCase
from os import urandom

from mock import Mock, patch, PropertyMock
from pyqrllib.pyqrllib import hstr2bin
from pyqryptonight.pyqryptonight import StringToUInt256

from qrl.generated import qrlstateinfo_pb2
from qrl.core import config
from qrl.core.misc import logger
from qrl.crypto.xmss import XMSS
from qrl.core.Block import Block
from qrl.core.ChainManager import ChainManager
from qrl.core.DifficultyTracker import DifficultyTracker
from qrl.core.GenesisBlock import GenesisBlock
from qrl.core.State import State
from qrl.core.txs.SlaveTransaction import SlaveTransaction
from qrl.core.txs.TransferTransaction import TransferTransaction
from tests.misc.helper import get_alice_xmss, get_bob_xmss, set_default_balance_size, set_qrl_dir, replacement_getTime

logger.initialize_default()

alice = get_alice_xmss()
bob = get_bob_xmss()


def ask_difficulty_tracker(difficulty: str):
    dt = DifficultyTracker()
    tmp_difficulty = StringToUInt256('2')
    tmp_target = dt.get_target(tmp_difficulty)
    return tmp_difficulty, tmp_target


def create_block(block_number, previous_block, miner_address):
    return Block.create(block_number=block_number,
                        prev_headerhash=previous_block.headerhash,
                        prev_timestamp=previous_block.timestamp,
                        transactions=[],
                        miner_address=miner_address
                        )


def create_m_block(block_number, previous_block, miner_address):
    mock_block = Mock(
        autospec=Block,
        name="Mock Block {}".format(block_number),
        block_number=block_number,
        prev_headerhash=previous_block.headerhash,
        prev_block_timestamp=previous_block.timestamp,
        transactions=[],
        miner_address=miner_address,
        headerhash="Mock Block {} {}".format(block_number, urandom(6)).encode(),
        timestamp=replacement_getTime()
    )
    mock_block.serialize.return_value = "Mock Block {}".format(block_number).encode()
    return mock_block


class TestChainManagerReal(TestCase):
    def setUp(self):
        with set_qrl_dir('no_data'):
            self.state = State()
            self.state.get_measurement = Mock(return_value=10000000)

            del GenesisBlock.instance  # Removing Singleton instance
            self.genesis_block = GenesisBlock()

            self.chain_manager = ChainManager(self.state)
            self.chain_manager._difficulty_tracker = Mock()

    def test_load(self):
        # load() has the following tasks:
        # Write Genesis Block into State immediately
        # Register block_number <-> blockhash mapping
        # Calculate difficulty Metadata for Genesis Block
        # Generate AddressStates from Genesis Block balances
        # Apply Genesis Block's transactions to the state
        self.chain_manager.load(self.genesis_block)
        block = self.state.get_block(GenesisBlock().headerhash)
        self.assertIsNotNone(block)

    def test_load_twice(self):
        self.chain_manager.load(self.genesis_block)

        # If we call load() a second time, it should check to see if we are forked and initiate recovery.
        # First we pretend we are not forked.
        self.state.get_fork_state = Mock(return_value=None)
        self.chain_manager._fork_recovery = Mock(name='mock _fork_recovery()')
        self.chain_manager.load(self.genesis_block)
        self.chain_manager._fork_recovery.assert_not_called()

        # If we pretend we are forked, it should call _fork_recovery().
        m_fork_state = Mock(autospec=qrlstateinfo_pb2.ForkState, initiator_headerhash=self.genesis_block.headerhash)
        self.state.get_fork_state.return_value = m_fork_state
        self.chain_manager.load(self.genesis_block)
        self.chain_manager._fork_recovery.assert_called_with(self.genesis_block, m_fork_state)

    @patch('qrl.core.misc.ntp.getTime')
    def test_simple_add_block(self, time_mock):
        # Simply test that adding a block on top of the genesis block works.
        self.chain_manager._difficulty_tracker.get.return_value = ask_difficulty_tracker('2')
        self.chain_manager.load(self.genesis_block)

        time_mock.return_value = 1615270948  # Very high to get an easy difficulty

        block_1 = Block.create(block_number=1,
                               prev_headerhash=self.genesis_block.headerhash,
                               prev_timestamp=self.genesis_block.timestamp,
                               transactions=[],
                               miner_address=alice.address)
        block_1.set_nonces(201, 0)

        # Uncomment only to determine the correct mining_nonce of above blocks
        # from qrl.core.PoWValidator import PoWValidator
        # while not PoWValidator().validate_mining_nonce(self.state, block_1.blockheader, False):
        #     block_1.set_nonces(block_1.mining_nonce + 1)
        #     print(block_1.mining_nonce)
        self.assertTrue(block_1.validate(self.chain_manager, {}))
        result = self.chain_manager.add_block(block_1)

        self.assertTrue(result)
        self.assertEqual(self.chain_manager.last_block, block_1)

    @set_default_balance_size()
    @patch('qrl.core.misc.ntp.getTime')
    def test_multi_output_transaction_add_block(self, time_mock):
        # Test that adding block with a multi-output Transaction updates everybody's balances correctly.
        self.chain_manager.load(self.genesis_block)

        extended_seed = "010300cebc4e25553afa0aab899f7838e59e18a48852fa9dfd5" \
                        "ae78278c371902aa9e6e9c1fa8a196d2dba0cbfd2f2d212d16c"
        random_xmss = XMSS.from_extended_seed(hstr2bin(extended_seed))

        transfer_transaction = TransferTransaction.create(addrs_to=[alice.address, random_xmss.address],
                                                          amounts=[40 * int(config.dev.shor_per_quanta),
                                                                   59 * int(config.dev.shor_per_quanta)],
                                                          fee=1 * config.dev.shor_per_quanta,
                                                          xmss_pk=bob.pk)
        transfer_transaction._data.nonce = 1
        transfer_transaction.sign(bob)

        time_mock.return_value = 1615270948  # Very high to get an easy difficulty

        block_1 = Block.create(block_number=1,
                               prev_headerhash=self.genesis_block.headerhash,
                               prev_timestamp=self.genesis_block.timestamp,
                               transactions=[transfer_transaction],
                               miner_address=alice.address)
        block_1.set_nonces(129, 0)

        # Uncomment only to determine the correct mining_nonce of above blocks
        # from qrl.core.PoWValidator import PoWValidator
        # while not PoWValidator().validate_mining_nonce(self.state, block_1.blockheader, False):
        #     block_1.set_nonces(block_1.mining_nonce + 1)
        #     print(block_1.mining_nonce)
        self.assertTrue(block_1.validate(self.chain_manager, {}))
        result = self.chain_manager.add_block(block_1)

        self.assertTrue(result)
        self.assertEqual(self.chain_manager.last_block, block_1)

        bob_addr_state = self.state.get_address_state(bob.address)
        alice_addr_state = self.state.get_address_state(alice.address)
        random_addr_state = self.state.get_address_state(random_xmss.address)

        self.assertEqual(bob_addr_state.balance, 0)
        self.assertEqual(alice_addr_state.balance,
                         140 * int(config.dev.shor_per_quanta) + block_1.block_reward + block_1.fee_reward)
        self.assertEqual(random_addr_state.balance, 159 * int(config.dev.shor_per_quanta))

    @patch("qrl.core.DifficultyTracker.DifficultyTracker.get")
    def test_add_block(self, mock_difficulty_tracker_get):
        """
        Add block_1 on genesis block (that registers Bob as Alice's slave)
        Add a competing fork_block on genesis block (without the SlaveTransaction)
        Add block_2 on fork_block (without the SlaveTransaction)
        Bob should be free from slavery now.
        """
        mock_difficulty_tracker_get.return_value = ask_difficulty_tracker('2')
        self.chain_manager.load(self.genesis_block)

        # Add block_1 on genesis block.
        slave_tx = SlaveTransaction.create(slave_pks=[bob.pk],
                                           access_types=[0],
                                           fee=0,
                                           xmss_pk=alice.pk)
        slave_tx.sign(alice)
        slave_tx._data.nonce = 1
        self.assertTrue(slave_tx.validate())
        with patch('qrl.core.misc.ntp.getTime') as time_mock:
            time_mock.return_value = 1615270948  # Very high to get an easy difficulty

            block_1 = Block.create(block_number=1,
                                   prev_headerhash=self.genesis_block.headerhash,
                                   prev_timestamp=self.genesis_block.timestamp,
                                   transactions=[slave_tx],
                                   miner_address=alice.address)
            block_1.set_nonces(2, 0)
            # Uncomment only to determine the correct mining_nonce of above blocks
            # from qrl.core.PoWValidator import PoWValidator
            # while not PoWValidator().validate_mining_nonce(self.state, block_1.blockheader, False):
            #     block_1.set_nonces(block_1.mining_nonce + 1)
            #     print(block_1.mining_nonce)
            self.assertTrue(block_1.validate(self.chain_manager, {}))
            result = self.chain_manager.add_block(block_1)

        self.assertTrue(result)
        self.assertEqual(self.chain_manager.last_block, block_1)

        # Yes, Bob is Alice's slave.
        alice_state = self.chain_manager.get_address_state(alice.address)
        self.assertEqual(len(alice_state.slave_pks_access_type), 1)
        self.assertTrue(str(bob.pk) in alice_state.slave_pks_access_type)

        # Add fork block on genesis block
        with patch('qrl.core.misc.ntp.getTime') as time_mock:
            time_mock.return_value = 1715270948  # Very high to get an easy difficulty
            fork_block = Block.create(block_number=1,
                                      prev_headerhash=self.genesis_block.headerhash,
                                      prev_timestamp=self.genesis_block.timestamp,
                                      transactions=[],
                                      miner_address=bob.address)

            fork_block.set_nonces(4, 0)
            # Uncomment only to determine the correct mining_nonce of above blocks
            # from qrl.core.PoWValidator import PoWValidator
            # while not PoWValidator().validate_mining_nonce(self.state, fork_block.blockheader, False):
            #     fork_block.set_nonces(fork_block.mining_nonce + 1)
            #     print(fork_block.mining_nonce)
            self.assertTrue(fork_block.validate(self.chain_manager, {}))
            result = self.chain_manager.add_block(fork_block)

        self.assertTrue(result)
        self.assertEqual(self.chain_manager.last_block, block_1)

        fork_block = self.state.get_block(fork_block.headerhash)
        self.assertIsNotNone(fork_block)

        # Add block_2 on fork_block.
        with patch('qrl.core.misc.ntp.getTime') as time_mock:
            time_mock.return_value = 1815270948  # Very high to get an easy difficulty
            block_2 = fork_block.create(block_number=2,
                                        prev_headerhash=fork_block.headerhash,
                                        prev_timestamp=fork_block.timestamp,
                                        transactions=[],
                                        miner_address=bob.address)

            block_2.set_nonces(1, 0)
            # Uncomment only to determine the correct mining_nonce of above blocks
            # from qrl.core.PoWValidator import PoWValidator
            # while not PoWValidator().validate_mining_nonce(state, block_2.blockheader, False):
            #     block_2.set_nonces(block_2.mining_nonce + 1, 0)
            #     print(block_2.mining_nonce)
            self.assertTrue(block_2.validate(self.chain_manager, {}))
            result = self.chain_manager.add_block(block_2)

        self.assertTrue(result)
        self.assertEqual(self.chain_manager.last_block.block_number, block_2.block_number)
        self.assertEqual(self.chain_manager.last_block.serialize(), block_2.serialize())

        # Now we are on the forked chain, Bob is no longer Alice's slave.
        alice_state = self.chain_manager.get_address_state(alice.address)
        self.assertFalse(str(bob.pk) in alice_state.slave_pks_access_type)

    @patch('qrl.core.misc.ntp.getTime', new=replacement_getTime)
    def test_get_headerhashes(self):
        block_1 = create_block(1, self.genesis_block, alice.address)
        block_2 = create_block(2, block_1, alice.address)

        self.chain_manager.load(self.genesis_block)
        self.chain_manager.add_block(block_1)
        self.chain_manager.add_block(block_2)

        node_header_hash = self.chain_manager.get_headerhashes(0)
        self.assertEqual(node_header_hash.headerhashes,
                         [self.genesis_block.headerhash, block_1.headerhash, block_2.headerhash])

    @patch('qrl.core.misc.ntp.getTime', new=replacement_getTime)
    def test_fork_recovery(self):
        # When the node finds that it has been on the slower chain all this time, it runs _fork_recovery() to _rollback
        # to an earlier state and switch to the longer chain.

        block_1 = create_block(1, self.genesis_block, alice.address)

        block_2 = create_block(2, block_1, alice.address)
        block_3 = create_block(3, block_2, alice.address)
        self.main_chain = [block_2, block_3]

        block_2_alt = create_block(2, block_1, bob.address)
        block_3_alt = create_block(3, block_2_alt, bob.address)
        block_4_alt = create_block(4, block_3_alt, bob.address)
        self.alt_chain = [block_2_alt, block_3_alt, block_4_alt]

        self.chain_manager.load(self.genesis_block)

        self.chain_manager.add_block(block_1)
        self.chain_manager.add_block(block_2)
        self.chain_manager.add_block(block_3)

        # Start adding the forked block to the chain manager. It accepts, and does nothing else.
        self.chain_manager.add_block(block_2_alt)
        self.assertEqual(self.chain_manager.last_block, block_3)

        # We lengthen the fork. Still, the chain manager stays on the first chain.
        self.chain_manager.add_block(block_3_alt)
        self.assertEqual(self.chain_manager.last_block, block_3)

        # When it is obvious that the fork is longer (has a higher cum. diff), the chain manager invokes _fork_recovery()
        # and switches over to the fork
        self.chain_manager.add_block(block_4_alt)
        self.assertEqual(self.chain_manager.last_block, block_4_alt)


class TestChainManager(TestCase):
    def setUp(self):
        self.state = Mock(autospec=State)
        self.state.get_measurement.return_value = 10000000

        del GenesisBlock.instance  # Removing Singleton instance
        self.genesis_block = GenesisBlock()

        self.chain_manager = ChainManager(self.state)
        self.chain_manager.tx_pool = Mock()
        self.chain_manager._difficulty_tracker = Mock()

    def test_fork_recovery_failed(self):
        # When switching to the longer chain fails, _fork_recovery() must _rollback and restore the shorter chain.
        # Mock out irrelevant functions
        self.chain_manager._update_block_number_mapping = Mock()

        # Switching to the new chain should fail!
        self.chain_manager.add_chain = Mock(return_value=False)
        self.chain_manager._rollback = Mock()

        block_1 = create_m_block(1, self.genesis_block, alice.address)
        block_2 = create_m_block(2, block_1, alice.address)

        block_1_alt = create_m_block(1, self.genesis_block, alice.address)
        block_2_alt = create_m_block(2, block_1_alt, alice.address)
        block_3_alt = create_m_block(3, block_2_alt, alice.address)

        fork_state = qrlstateinfo_pb2.ForkState(
            initiator_headerhash=block_3_alt.headerhash,
            fork_point_headerhash=self.genesis_block.headerhash,
            old_mainchain_hash_path=[b.headerhash for b in [block_1, block_2]],
            new_mainchain_hash_path=[b.headerhash for b in [block_1_alt, block_2_alt, block_3_alt]]
        )

        # _fork_recovery() will _rollback() to the genesis block and go on the longer chain.
        # At this point, _rollback() should return the old hash path as a backup
        # in case switching to the longer chain fails.
        self.chain_manager._rollback.return_value = [block_2.headerhash, block_1.headerhash]

        self.chain_manager._fork_recovery(block_3_alt, fork_state)

        # _fork_recovery() should have _rollback()ed when trying to switch to the longer chain
        self.chain_manager._rollback.assert_any_call(self.genesis_block.headerhash, fork_state)
        # _fork_recovery() should have _rollback()ed to the genesis block when trying to restore the shorter chain
        self.chain_manager._rollback.assert_called_with(self.genesis_block.headerhash)

    def test_fork_recovery_rollbacked_already(self):
        # Switching to the new chain works, but test that if the _rollback() has already happened, _fork_recovery() does
        # not hiccup

        # Mock out irrelevant functions
        self.chain_manager._update_block_number_mapping = Mock()

        # Switching to the new chain should succeed!
        self.chain_manager.add_chain = Mock(return_value=True)
        self.chain_manager._rollback = Mock()

        block_1 = create_m_block(1, self.genesis_block, alice.address)
        block_2 = create_m_block(2, block_1, alice.address)

        block_1_alt = create_m_block(1, self.genesis_block, alice.address)
        block_2_alt = create_m_block(2, block_1_alt, alice.address)
        block_3_alt = create_m_block(3, block_2_alt, alice.address)

        fork_state = qrlstateinfo_pb2.ForkState(
            initiator_headerhash=block_3_alt.headerhash,
            fork_point_headerhash=self.genesis_block.headerhash,
            old_mainchain_hash_path=[b.headerhash for b in [block_1, block_2]],
            new_mainchain_hash_path=[b.headerhash for b in [block_1_alt, block_2_alt, block_3_alt]]
        )

        # State.get_block() should say that we are already on block_1_alt
        self.chain_manager._state.get_block.return_value = block_1_alt

        # _fork_recovery() will not call _rollback(), because it has already happened.
        self.chain_manager._fork_recovery(block_3_alt, fork_state)

        # _fork_recovery() should have _rollback()ed when trying to switch to the longer chain
        self.chain_manager._rollback.assert_not_called()

    @patch('qrl.core.config')
    @patch('qrl.core.ChainManager.ChainManager.height', new_callable=PropertyMock)
    def test_add_block_doesnt_add_blocks_beyond_reorg_limit(self, m_height, m_config):
        # If we are at height 40000, what's the use of adding a block that's height 1? Simply ignore that block.
        m_config.dev.reorg_limit = 22000
        m_height.return_value = 40000
        block_1 = create_m_block(1, self.genesis_block, alice.address)

        ans = self.chain_manager.add_block(block_1)
        self.assertFalse(ans)

    def test_add_block_refuses_to_add_too_large_blocks(self):
        # State.get_block_size_limit() calculates how large each Block should be from the last 10 confirmed blocks.
        self.state.get_block_size_limit.return_value = 5000000
        block_1 = create_m_block(1, self.genesis_block, alice.address)
        block_1.size = 6000000

        ans = self.chain_manager.add_block(block_1)
        self.assertFalse(ans)

    def test_get_fork_point_failure_modes(self):
        block_0 = create_m_block(0, Mock(headerhash=b'Fake Genesis', timestamp=replacement_getTime()), alice.address)
        block_1 = create_m_block(1, block_0, alice.address)
        block_2 = create_m_block(2, block_1, alice.address)

        # If _get_fork_point() ever reaches block_number 0, that means the genesis block is different!
        # Mock self.state leads us back to block_0
        self.state.get_block.side_effect = [block_2, block_1, block_0]

        with self.assertRaises(Exception):
            self.chain_manager._get_fork_point(block_2)

        # If _get_fork_point() cannot find a particular block while walking back to the fork point, something has gone
        # very wrong
        # Mock self.state leads us back through a broken chain
        self.state.get_block.side_effect = [block_2, None, block_0]

        with self.assertRaises(Exception):
            self.chain_manager._get_fork_point(block_2)

    def test_apply_block_fails_if_state_changes_fail(self):
        # ChainManager._apply_block() should fail if Block.apply_state_changes() fails.
        block = create_m_block(50, self.genesis_block, alice.address)
        block.apply_state_changes.return_value = False

        ans = self.chain_manager._apply_block(block, [])
        self.assertFalse(ans)

    def test_try_branch_add_block_fails_if_apply_block_fails(self):
        # ChainManager._try_branch_add_block() should fail if ChainManager._apply_block() fails
        self.chain_manager._apply_block = Mock(return_value=False)

        block = create_m_block(50, self.genesis_block, alice.address)

        block_added, fork_flag = self.chain_manager._try_branch_add_block(block, [])
        self.assertFalse(block_added)
        self.assertFalse(fork_flag)

    def test_add_chain_fails_if_fork_recovery_didnt_complete_successfully(self):
        block_1 = create_m_block(1, self.genesis_block, alice.address)
        block_2 = create_m_block(2, block_1, alice.address)

        block_1_alt = create_m_block(1, self.genesis_block, alice.address)
        block_2_alt = create_m_block(2, block_1_alt, alice.address)
        block_3_alt = create_m_block(3, block_2_alt, alice.address)

        fork_state = qrlstateinfo_pb2.ForkState(
            initiator_headerhash=block_3_alt.headerhash,
            fork_point_headerhash=self.genesis_block.headerhash,
            old_mainchain_hash_path=[b.headerhash for b in [block_1, block_2]],
            new_mainchain_hash_path=[b.headerhash for b in [block_1_alt, block_2_alt, block_3_alt]]
        )
        # We want to add_chain(block_*_alt chain), but we're still on block_1 (we should have rolled back to genesis)
        self.chain_manager._last_block = block_1
        ans = self.chain_manager.add_chain([block_1_alt.headerhash, block_2_alt.headerhash], fork_state)
        self.assertFalse(ans)

    def test_add_chain_fails_if_apply_block_fails(self):
        block_1 = create_m_block(1, self.genesis_block, alice.address)
        block_2 = create_m_block(2, block_1, alice.address)

        block_1_alt = create_m_block(1, self.genesis_block, alice.address)
        block_2_alt = create_m_block(2, block_1_alt, alice.address)
        block_3_alt = create_m_block(3, block_2_alt, alice.address)

        fork_state = qrlstateinfo_pb2.ForkState(
            initiator_headerhash=block_3_alt.headerhash,
            fork_point_headerhash=self.genesis_block.headerhash,
            old_mainchain_hash_path=[b.headerhash for b in [block_1, block_2]],
            new_mainchain_hash_path=[b.headerhash for b in [block_1_alt, block_2_alt, block_3_alt]]
        )

        # we want to add_chain(block_*_alt chain), but for some reason applying a Block to the State didn't work.
        self.chain_manager._apply_block = Mock(return_value=False)
        ans = self.chain_manager.add_chain([block_1_alt.headerhash, block_2_alt.headerhash], fork_state)
        self.assertFalse(ans)

    def test_get_transaction(self):
        # get_transaction() is simply a wrapper for State.get_tx_metadata
        self.chain_manager.get_tx_metadata(b'txhash')
        self.state.get_tx_metadata.assert_called_once()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    This file was generated with PyScaffold 3.0.2.
    PyScaffold helps you to put up the scaffold of your new Python project.
    Learn more under: http://pyscaffold.org/
"""

import sys
from setuptools import setup
import versioneer

# Add here console scripts and other entry points in ini-style format
entry_points = """
[console_scripts]
    start_qrl = qrl.main:main
    qrl_start = qrl.main:main
    qrl = qrl.cli:main
    qrl_grpc_proxy = qrl.grpcProxy:main
    qrl_measure = qrl.measure:main
    qrl_walletd = qrl.daemon.walletd:main
    qrl_generate_genesis = qrl.tools.generate_genesis:main
"""


def setup_package():
    needs_sphinx = {'build_sphinx', 'upload_docs'}.intersection(sys.argv)
    sphinx = ['sphinx'] if needs_sphinx else []
    setup(setup_requires=['pyscaffold>=3.0a0,<3.1a0'] + sphinx,
          entry_points=entry_points,
          version=versioneer.get_version(),
          cmdclass=versioneer.get_cmdclass(),
          use_pyscaffold=True)


if __name__ == "__main__":
    setup_package()



from unittest import TestCase

import simplejson as json
from mock import patch, PropertyMock, Mock
from pyqrllib.pyqrllib import bin2hstr

from qrl.core.misc import logger
from qrl.core.AddressState import AddressState
from qrl.core.ChainManager import ChainManager
from qrl.core.TransactionInfo import TransactionInfo
from qrl.core.txs.Transaction import Transaction
from qrl.core.txs.TransferTransaction import TransferTransaction
from tests.core.txs.testdata import test_json_Simple, test_signature_Simple
from tests.misc.helper import get_alice_xmss, get_bob_xmss, get_slave_xmss, replacement_getTime

logger.initialize_default()


@patch('qrl.core.txs.Transaction.logger')
class TestSimpleTransaction(TestCase):

    def __init__(self, *args, **kwargs):
        super(TestSimpleTransaction, self).__init__(*args, **kwargs)
        self.alice = get_alice_xmss()
        self.bob = get_bob_xmss()
        self.slave = get_slave_xmss()

        self.alice.set_ots_index(10)
        self.maxDiff = None

    def setUp(self):
        self.tx = TransferTransaction.create(
            addrs_to=[self.bob.address],
            amounts=[100],
            fee=1,
            xmss_pk=self.alice.pk
        )

    def test_create(self, m_logger):
        # Alice sending coins to Bob
        tx = TransferTransaction.create(addrs_to=[self.bob.address],
                                        amounts=[100],
                                        fee=1,
                                        xmss_pk=self.alice.pk)
        self.assertTrue(tx)

    def test_create_negative_amount(self, m_logger):
        with self.assertRaises(ValueError):
            TransferTransaction.create(addrs_to=[self.bob.address],
                                       amounts=[-100],
                                       fee=1,
                                       xmss_pk=self.alice.pk)

    def test_create_negative_fee(self, m_logger):
        with self.assertRaises(ValueError):
            TransferTransaction.create(addrs_to=[self.bob.address],
                                       amounts=[-100],
                                       fee=-1,
                                       xmss_pk=self.alice.pk)

    def test_to_json(self, m_logger):
        tx = TransferTransaction.create(addrs_to=[self.bob.address],
                                        amounts=[100],
                                        fee=1,
                                        xmss_pk=self.alice.pk)
        txjson = tx.to_json()

        self.assertEqual(json.loads(test_json_Simple), json.loads(txjson))

    def test_from_json(self, m_logger):
        tx = Transaction.from_json(test_json_Simple)
        tx.sign(self.alice)
        self.assertIsInstance(tx, TransferTransaction)

        # Test that common Transaction components were copied over.
        self.assertEqual(0, tx.nonce)
        self.assertEqual('010300a1da274e68c88b0ccf448e0b1916fa789b01eb2ed4e9ad565ce264c9390782a9c61ac02f',
                         bin2hstr(tx.addr_from))
        self.assertEqual('01030038ea6375069f8272cc1a6601b3c76c21519455603d370036b97c779ada356'
                         '5854e3983bd564298c49ae2e7fa6e28d4b954d8cd59398f1225b08d6144854aee0e',
                         bin2hstr(tx.PK))
        self.assertEqual('554f546305d4aed6ec71c759942b721b904ab9d65eeac3c954c08c652181c4e8', bin2hstr(tx.txhash))
        self.assertEqual(10, tx.ots_key)

        self.assertEqual(test_signature_Simple, bin2hstr(tx.signature))

        # Test that specific content was copied over.
        self.assertEqual('0103001d65d7e59aed5efbeae64246e0f3184d7c42411421eb385ba30f2c1c005a85ebc4419cfd',
                         bin2hstr(tx.addrs_to[0]))
        self.assertEqual(100, tx.total_amount)
        self.assertEqual(1, tx.fee)

    def test_validate_tx(self, m_logger):
        # If we change amount, fee, addr_from, addr_to, (maybe include xmss stuff) txhash should change.
        # Here we use the tx already defined in setUp() for convenience.
        # We must sign the tx before validation will work.
        self.tx.sign(self.alice)

        # We have not touched the tx: validation should pass.
        self.assertTrue(self.tx.validate_or_raise())

    def test_validate_tx2(self, m_logger):
        tx = TransferTransaction.create(
            addrs_to=[self.bob.address],
            amounts=[100],
            fee=1,
            xmss_pk=self.alice.pk
        )
        tx.sign(self.alice)

        self.assertTrue(tx.validate_or_raise())

        tx._data.transaction_hash = b'abc'

        # Should fail, as we have modified with invalid transaction_hash
        with self.assertRaises(ValueError):
            tx.validate_or_raise()

    def test_validate_tx3(self, m_logger):
        tx = TransferTransaction.create(
            addrs_to=[self.bob.address],
            amounts=[100],
            fee=1,
            xmss_pk=self.alice.pk
        )
        tx.sign(self.alice)
        tx._data.signature = tx.signature * 4183 + tx.signature[0:104]
        tx._data.transaction_hash = tx.generate_txhash()

        with self.assertRaises(ValueError):
            tx.validate_or_raise()

    @patch('qrl.core.txs.Transaction.config')
    def test_validate_tx_invalid(self, m_config, m_logger):
        # Test all the things that could make a TransferTransaction invalid
        self.tx.sign(self.alice)

        # Validation in creation, Protobuf, type conversion etc. gets in our way all the time!
        # So to get dirty data to the validate() function, we need PropertyMocks
        with patch('qrl.core.txs.TransferTransaction.TransferTransaction.amounts',
                   new_callable=PropertyMock) as m_amounts:
            # TX amount of 0 shouldn't be allowed.
            m_amounts.return_value = [0]
            with self.assertRaises(ValueError):
                self.tx.validate_or_raise()
        with patch('qrl.core.txs.TransferTransaction.TransferTransaction.fee', new_callable=PropertyMock) as m_fee:
            m_fee.return_value = -1
            with self.assertRaises(ValueError):
                self.tx.validate_or_raise()
        with patch('qrl.core.txs.TransferTransaction.TransferTransaction.addrs_to',
                   new_callable=PropertyMock) as m_addrs_to:
            with patch('qrl.core.txs.TransferTransaction.TransferTransaction.amounts',
                       new_callable=PropertyMock) as m_amounts:
                # Validation could fail because len(m_addrs_to) != len(m_amounts),
                # or if len(m_addrs_to) > transaction_multi_output_limit.
                # This second patch level is to make sure the only the latter case happens.
                m_amounts = [100, 100, 100, 100]
                m_config.dev.transaction_multi_output_limit = 3
                m_addrs_to.return_value = [2, 2, 2, 2]
                with self.assertRaises(ValueError):
                    self.tx.validate_or_raise()
        with patch('qrl.core.txs.TransferTransaction.TransferTransaction.addrs_to',
                   new_callable=PropertyMock) as m_addrs_to:
            # len(addrs_to) must equal len(amounts)
            m_addrs_to.return_value = [2, 2]
            with self.assertRaises(ValueError):
                self.tx.validate_or_raise()
        with patch('qrl.core.txs.TransferTransaction.TransferTransaction.addr_from',
                   new_callable=PropertyMock) as m_addr_from:
            m_addr_from.return_value = b'If this isnt invalid Ill eat my shoe'
            with self.assertRaises(ValueError):
                self.tx.validate_or_raise()
        with patch('qrl.core.txs.TransferTransaction.TransferTransaction.addrs_to',
                   new_callable=PropertyMock) as m_addrs_to:
            with patch('qrl.core.txs.TransferTransaction.TransferTransaction.amounts',
                       new_callable=PropertyMock) as m_amounts:
                m_amounts.return_value = [100, 100]
                m_addrs_to.return_value = [self.bob.address, b'If this isnt invalid Ill eat my shoe']
                with self.assertRaises(ValueError):
                    self.tx.validate_or_raise()

    def test_validate_extended(self, m_logger):
        """
        validate_extended() handles these parts of the validation:
        1. Master/slave
        2. balance, amount + fee from AddressState
        3. OTS key reuse from AddressState
        :return:
        """
        m_addr_state = Mock(autospec=AddressState, balance=200)
        m_addr_from_pk_state = Mock(autospec=AddressState)
        m_addr_from_pk_state.ots_key_reuse.return_value = False

        self.tx.validate_slave = Mock(autospec=Transaction.validate_slave, return_value=True)

        self.tx.sign(self.alice)

        result = self.tx.validate_extended(m_addr_state, m_addr_from_pk_state)

        self.assertTrue(result)

        # Suppose there was ots key reuse. The function should then return false.
        m_addr_from_pk_state.ots_key_reuse.return_value = True
        result = self.tx.validate_extended(m_addr_state, m_addr_from_pk_state)
        self.assertFalse(result)

        # Reset conditions from above
        m_addr_from_pk_state.ots_key_reuse.return_value = False
        # Suppose the slave XMSS address does not have permission for this type of Transaction. It should return False.
        self.tx.validate_slave.return_value = False
        result = self.tx.validate_extended(m_addr_state, m_addr_from_pk_state)
        self.assertFalse(result)

        # Reset conditions from above
        self.tx.validate_slave.return_value = True
        # Suppose the address doesn't have enough coins.
        m_addr_state.balance = 99
        result = self.tx.validate_extended(m_addr_state, m_addr_from_pk_state)
        self.assertFalse(result)

    def test_validate_transaction_pool(self, m_logger):
        """
        Two TransferTransactions. Although they're the same, they are signed with different OTS indexes.
        Therefore they should not conflict when they are both in the TransactionPool.
        :return:
        """
        tx = self.tx
        tx2 = TransferTransaction.create(
            addrs_to=[self.bob.address],
            amounts=[100],
            fee=1,
            xmss_pk=self.alice.pk
        )
        tx.sign(self.alice)
        tx2.sign(self.alice)
        tx_info = Mock(autospec=TransactionInfo, transaction=tx)
        tx2_info = Mock(autospec=TransactionInfo, transaction=tx2)
        transaction_pool = [(replacement_getTime(), tx_info), (replacement_getTime(), tx2_info)]
        result = tx.validate_transaction_pool(transaction_pool)
        self.assertTrue(result)

    def test_validate_transaction_pool_reusing_ots_index(self, m_logger):
        """
        Two different TransferTransactions. They are signed with the same OTS indexe, from the same public key.
        Therefore they should conflict.
        :return:
        """
        tx = self.tx
        tx2 = TransferTransaction.create(
            addrs_to=[self.bob.address],
            amounts=[100],
            fee=5,
            xmss_pk=self.alice.pk
        )

        # alice_clone's OTS index is still at 10, while self.alice will be at 11 after signing.
        alice_clone = get_alice_xmss()
        alice_clone.set_ots_index(10)
        tx.sign(self.alice)
        tx2.sign(alice_clone)

        tx_info = Mock(autospec=TransactionInfo, transaction=tx)
        tx2_info = Mock(autospec=TransactionInfo, transaction=tx2)
        transaction_pool = [(replacement_getTime(), tx_info), (replacement_getTime(), tx2_info)]

        result = tx.validate_transaction_pool(transaction_pool)
        self.assertFalse(result)

    def test_validate_transaction_pool_different_pk_same_ots_index(self, m_logger):
        """
        Two TransferTransactions. They are signed with the same OTS indexes, but from different public keys.
        Therefore they should NOT conflict.
        :return:
        """
        tx = self.tx
        tx2 = TransferTransaction.create(
            addrs_to=[self.bob.address],
            amounts=[100],
            fee=1,
            xmss_pk=self.bob.pk
        )

        tx.sign(self.alice)
        tx2.sign(self.bob)

        tx_info = Mock(autospec=TransactionInfo, transaction=tx)
        tx2_info = Mock(autospec=TransactionInfo, transaction=tx2)
        transaction_pool = [(replacement_getTime(), tx_info), (replacement_getTime(), tx2_info)]

        result = tx.validate_transaction_pool(transaction_pool)
        self.assertTrue(result)

    def test_apply_state_changes(self, m_logger):
        """
        apply_state_changes() is the part that actually updates everybody's balances.
        Then it forwards the addresses_state to _apply_state_changes_for_PK(), which updates everybody's addresses's
        nonce, OTS key index, and associated TX hashes
        If there is no AddressState for a particular Address, nothing is done.
        """
        addresses_state = {
            self.alice.address: Mock(autospec=AddressState, name='alice AddressState', transaction_hashes=[],
                                     balance=200),
            self.bob.address: Mock(autospec=AddressState, name='bob AddressState', transaction_hashes=[], balance=0),
            self.slave.address: Mock(autospec=AddressState, name='slave AddressState', transaction_hashes=[], balance=0)
        }
        self.tx._apply_state_changes_for_PK = Mock(autospec=TransferTransaction._apply_state_changes_for_PK)

        self.tx.apply_state_changes(addresses_state)

        # Now Alice should have 99 coins left (200 - 100 - 1) and Bob should have 100 coins.
        self.assertEqual(99, addresses_state[self.alice.address].balance)
        self.assertEqual(100, addresses_state[self.bob.address].balance)
        self.tx._apply_state_changes_for_PK.assert_called_once()

        # If there are no AddressStates related to the Addresses in this transaction, do nothing.
        self.tx._apply_state_changes_for_PK.reset_mock()
        addresses_state_dummy = {
            b'a': 'ABC',
            b'b': 'DEF'
        }
        self.tx.apply_state_changes(addresses_state_dummy)
        self.assertEqual(addresses_state_dummy, {b'a': 'ABC', b'b': 'DEF'})
        self.tx._apply_state_changes_for_PK.assert_called_once()

    def test_apply_state_changes_tx_sends_to_self(self, m_logger):
        """
        If you send coins to yourself, you should only lose the fee for the Transaction.
        """
        addresses_state = {
            self.alice.address: Mock(autospec=AddressState, name='alice AddressState', transaction_hashes=[],
                                     balance=200),
            self.bob.address: Mock(autospec=AddressState, name='bob AddressState', transaction_hashes=[], balance=0),
            self.slave.address: Mock(autospec=AddressState, name='slave AddressState', transaction_hashes=[], balance=0)
        }
        tx = TransferTransaction.create(
            addrs_to=[self.alice.address],
            amounts=[100],
            fee=1,
            xmss_pk=self.alice.pk
        )
        tx._apply_state_changes_for_PK = Mock(autospec=TransferTransaction._revert_state_changes_for_PK)

        tx.apply_state_changes(addresses_state)

        self.assertEqual(199, addresses_state[self.alice.address].balance)
        self.assertIn(tx.txhash, addresses_state[self.alice.address].transaction_hashes)

    def test_apply_state_changes_multi_send(self, m_logger):
        """
        Test that apply_state_changes() also works with multiple recipients.
        """
        addresses_state = {
            self.alice.address: Mock(autospec=AddressState, name='alice AddressState', transaction_hashes=[],
                                     balance=200),
            self.bob.address: Mock(autospec=AddressState, name='bob AddressState', transaction_hashes=[], balance=0),
            self.slave.address: Mock(autospec=AddressState, name='slave AddressState', transaction_hashes=[], balance=0)
        }

        tx_multisend = TransferTransaction.create(
            addrs_to=[self.bob.address, self.slave.address],
            amounts=[20, 20],
            fee=1,
            xmss_pk=self.alice.pk
        )
        tx_multisend._apply_state_changes_for_PK = Mock(autospec=TransferTransaction._apply_state_changes_for_PK)

        tx_multisend.apply_state_changes(addresses_state)

        # Now Alice should have 159 coins left (200 - 20 - 20 - 1) and Bob should have 100 coins.
        self.assertEqual(159, addresses_state[self.alice.address].balance)
        self.assertEqual(20, addresses_state[self.bob.address].balance)
        self.assertEqual(20, addresses_state[self.slave.address].balance)
        tx_multisend._apply_state_changes_for_PK.assert_called_once()

    def test_apply_state_changes_for_PK(self, m_logger):
        """
        This updates the node's AddressState database with which OTS index a particular address should be on, and what
        tx hashes is this address associated with.
        Curiously enough, if the TX was signed by a master XMSS tree, it doesn't add this tx's txhash to the list of
        txs that address is associated with.
        :return:
        """
        addr_state = {
            self.alice.address: Mock(autospec=AddressState)
        }
        old_ots_index = self.alice.ots_index
        self.tx.sign(self.alice)
        self.tx._apply_state_changes_for_PK(addr_state)

        addr_state[self.alice.address].increase_nonce.assert_called_once()
        addr_state[self.alice.address].set_ots_key.assert_called_once_with(old_ots_index)

    def test_apply_state_changes_for_PK_master_slave_XMSS(self, m_logger):
        """
        If the TX was signed by a slave XMSS, the slave XMSS's AddressState should be updated (not the master's).
        :return:
        """
        tx = TransferTransaction.create(
            addrs_to=[self.bob.address],
            amounts=[100],
            fee=1,
            xmss_pk=self.slave.pk,
            master_addr=self.alice.address
        )
        addr_state = {
            self.alice.address: Mock(autospec=AddressState, name='alice AddressState'),
            self.slave.address: Mock(autospec=AddressState, name='slave AddressState')
        }
        old_ots_index = self.slave.ots_index
        tx.sign(self.slave)
        tx._apply_state_changes_for_PK(addr_state)

        addr_state[self.slave.address].increase_nonce.assert_called_once()
        addr_state[self.slave.address].set_ots_key.assert_called_once_with(old_ots_index)
        addr_state[self.slave.address].transaction_hashes.append.assert_called_once()

    def test_revert_state_changes(self, m_logger):
        """
        Alice has sent 100 coins to Bob, using 1 as Transaction fee. Now we need to undo this.
        """
        addresses_state = {
            self.alice.address: Mock(autospec=AddressState, name='alice AddressState',
                                     transaction_hashes=[self.tx.txhash],
                                     balance=99),
            self.bob.address: Mock(autospec=AddressState, name='bob AddressState', transaction_hashes=[self.tx.txhash],
                                   balance=100),
            self.slave.address: Mock(autospec=AddressState, name='slave AddressState', transaction_hashes=[], balance=0)
        }

        unused_chain_manager_mock = Mock(autospec=ChainManager, name='unused ChainManager')
        self.tx._revert_state_changes_for_PK = Mock(autospec=TransferTransaction._revert_state_changes_for_PK)

        self.tx.revert_state_changes(addresses_state, unused_chain_manager_mock)

        self.assertEqual(200, addresses_state[self.alice.address].balance)
        self.assertEqual(0, addresses_state[self.bob.address].balance)
        self.assertEqual([], addresses_state[self.alice.address].transaction_hashes)
        self.assertEqual([], addresses_state[self.bob.address].transaction_hashes)

        self.tx._revert_state_changes_for_PK.assert_called_once()

        # If there are no AddressStates related to the Addresses in this transaction, do nothing.
        self.tx._revert_state_changes_for_PK.reset_mock()
        addresses_state_dummy = {
            b'a': 'ABC',
            b'b': 'DEF'
        }
        self.tx.revert_state_changes(addresses_state_dummy, unused_chain_manager_mock)
        self.assertEqual(addresses_state_dummy, {b'a': 'ABC', b'b': 'DEF'})
        self.tx._revert_state_changes_for_PK.assert_called_once()

    def test_revert_state_changes_multi_send(self, m_logger):
        """
        Alice has sent 20 coins to Bob and Slave each, using 1 as Transaction fee. Now we need to undo this.
        """
        addresses_state = {
            self.alice.address: Mock(autospec=AddressState, name='alice AddressState',
                                     transaction_hashes=[self.tx.txhash],
                                     balance=159),
            self.bob.address: Mock(autospec=AddressState, name='bob AddressState', transaction_hashes=[self.tx.txhash],
                                   balance=20),
            self.slave.address: Mock(autospec=AddressState, name='slave AddressState',
                                     transaction_hashes=[self.tx.txhash], balance=20)
        }
        unused_chain_manager_mock = Mock(autospec=ChainManager, name='unused ChainManager')

        tx_multisend = TransferTransaction.create(
            addrs_to=[self.bob.address, self.slave.address],
            amounts=[20, 20],
            fee=1,
            xmss_pk=self.alice.pk
        )
        tx_multisend._revert_state_changes_for_PK = Mock(autospec=TransferTransaction._revert_state_changes_for_PK)

        tx_multisend.revert_state_changes(addresses_state, unused_chain_manager_mock)

        self.assertEqual(200, addresses_state[self.alice.address].balance)
        self.assertEqual(0, addresses_state[self.bob.address].balance)
        self.assertEqual(0, addresses_state[self.slave.address].balance)
        self.assertEqual([], addresses_state[self.alice.address].transaction_hashes)
        self.assertEqual([], addresses_state[self.bob.address].transaction_hashes)
        self.assertEqual([], addresses_state[self.slave.address].transaction_hashes)

        tx_multisend._revert_state_changes_for_PK.assert_called_once()

    def test_revert_state_changes_tx_sends_to_self(self, m_logger):
        """
        Alice sent coins to herself, but she still lost the Transaction fee. Undo this.
        """
        addresses_state = {
            self.alice.address: Mock(autospec=AddressState, name='alice AddressState',
                                     transaction_hashes=[self.tx.txhash],
                                     balance=199),
            self.bob.address: Mock(autospec=AddressState, name='bob AddressState', transaction_hashes=[],
                                   balance=0),
            self.slave.address: Mock(autospec=AddressState, name='slave AddressState', transaction_hashes=[], balance=0)
        }

        unused_chain_manager_mock = Mock(autospec=ChainManager, name='unused ChainManager')

        tx = TransferTransaction.create(
            addrs_to=[self.alice.address],
            amounts=[100],
            fee=1,
            xmss_pk=self.alice.pk
        )

        tx._revert_state_changes_for_PK = Mock(autospec=TransferTransaction._revert_state_changes_for_PK)

        tx.revert_state_changes(addresses_state, unused_chain_manager_mock)

        self.assertEqual(200, addresses_state[self.alice.address].balance)
        self.assertEqual(0, addresses_state[self.bob.address].balance)
        self.assertEqual([], addresses_state[self.alice.address].transaction_hashes)
        self.assertEqual([], addresses_state[self.bob.address].transaction_hashes)

        tx._revert_state_changes_for_PK.assert_called_once()

    def test_revert_state_changes_for_PK(self, m_logger):
        """
        This is just an undo function.
        :return:
        """
        tx = TransferTransaction.create(
            addrs_to=[self.bob.address],
            amounts=[100],
            fee=1,
            xmss_pk=self.alice.pk
        )
        addr_state = {
            self.alice.address: Mock(autospec=AddressState)
        }
        tx.sign(self.alice)
        tx._revert_state_changes_for_PK(addr_state, Mock(name='unused State Mock'))

        addr_state[self.alice.address].decrease_nonce.assert_called_once()
        addr_state[self.alice.address].unset_ots_key.assert_called_once()

    def test_revert_state_changes_for_PK_master_slave_XMSS(self, m_logger):
        tx = TransferTransaction.create(
            addrs_to=[self.bob.address],
            amounts=[100],
            fee=1,
            xmss_pk=self.slave.pk,
            master_addr=self.alice.address
        )
        addr_state = {
            self.alice.address: Mock(autospec=AddressState, name='alice AddressState'),
            self.slave.address: Mock(autospec=AddressState, name='slave AddressState')
        }
        tx.sign(self.slave)
        tx._revert_state_changes_for_PK(addr_state, Mock(name='unused State Mock'))

        addr_state[self.slave.address].decrease_nonce.assert_called_once()
        addr_state[self.slave.address].unset_ots_key.assert_called_once()
        addr_state[self.slave.address].transaction_hashes.remove.assert_called_once()

    def test_affected_address(self, m_logger):
        # The default transaction params involve only two addresses.
        affected_addresses = set()
        self.tx.set_affected_address(affected_addresses)
        self.assertEqual(2, len(affected_addresses))

        # This transaction should involve 3 addresses.
        affected_addresses = set()
        tx = TransferTransaction.create(
            addrs_to=[self.bob.address, self.slave.address],
            amounts=[100, 100],
            fee=1,
            xmss_pk=self.alice.pk
        )
        tx.set_affected_address(affected_addresses)
        self.assertEqual(3, len(affected_addresses))



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from collections import OrderedDict
from unittest import TestCase

from mock import Mock, patch
from pyqrllib.pyqrllib import hstr2bin
from pyqryptonight.pyqryptonight import StringToUInt256
from twisted.internet import reactor

from qrl.core import config
from qrl.core.Block import Block
from qrl.core.State import State
from qrl.core.ChainManager import ChainManager
from qrl.core.Message import Message
from qrl.core.MessageRequest import MessageRequest
from qrl.core.misc import logger
from qrl.core.node import POW
from qrl.core.p2p.IPMetadata import IPMetadata
from qrl.core.p2p.p2pPeerManager import P2PPeerManager
from qrl.core.p2p.p2pfactory import P2PFactory
from qrl.core.p2p.p2pprotocol import P2PProtocol
from qrl.core.qrlnode import QRLNode
from qrl.core.txs.MessageTransaction import MessageTransaction
from qrl.core.txs.SlaveTransaction import SlaveTransaction
from qrl.core.txs.TokenTransaction import TokenTransaction
from qrl.core.txs.TransferTokenTransaction import TransferTokenTransaction
from qrl.core.txs.TransferTransaction import TransferTransaction
from qrl.generated import qrl_pb2, qrllegacy_pb2
from tests.misc.helper import replacement_getTime

logger.initialize_default()


def bhstr2bin(a_string: str) -> bytes:
    return bytes(hstr2bin(a_string))


def make_message(**kwargs):
    return qrllegacy_pb2.LegacyMessage(**kwargs)


def make_address(ip, port=config.user.p2p_public_port):
    return '{}:{}'.format(ip, port)


@patch('qrl.core.misc.ntp.getTime', new=replacement_getTime)
@patch('qrl.core.p2p.p2pfactory.logger', autospec=logger, name="Mock Logger")
@patch('qrl.core.p2p.p2pfactory.reactor', autospec=reactor, name="My Mock Reactor")
class TestP2PFactory(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestP2PFactory, self).__init__(*args, **kwargs)

    def setUp(self):
        self.m_qrlnode = Mock(autospec=QRLNode, name='Fake QRLNode')
        self.m_qrlnode.peer_manager = Mock(autospec=P2PPeerManager, name='Fake PeerManager')
        self.m_qrlnode.peer_manager.is_banned.return_value = False

        self.channel_1 = Mock(autospec=P2PProtocol,
                              name='mock Channel 1',
                              peer=IPMetadata('1.1.1.1', config.user.p2p_public_port))
        self.channel_2 = Mock(autospec=P2PProtocol,
                              name='mock Channel 2',
                              peer=IPMetadata('2.2.2.2', config.user.p2p_public_port))
        self.channel_3 = Mock(autospec=P2PProtocol,
                              name='mock Channel 3',
                              peer=IPMetadata('3.3.3.3', config.user.p2p_public_port))

        self.factory = P2PFactory(chain_manager=ChainManager(state=Mock(autospec=State)), sync_state=None, qrl_node=self.m_qrlnode)
        self.factory.pow = Mock(autospec=POW)

        self.factory.add_connection(self.channel_1)
        self.factory.add_connection(self.channel_2)
        self.factory.add_connection(self.channel_3)

    def tearDown(self):
        self.factory.remove_connection(self.channel_1)
        self.factory.remove_connection(self.channel_2)
        self.factory.remove_connection(self.channel_3)

    def test_create_factory(self, m_reactor, m_logger):
        factory = P2PFactory(chain_manager=ChainManager(None), sync_state=None, qrl_node=None)
        self.assertEqual(P2PProtocol, factory.protocol,
                         "Factory has not been assigned the expected protocol")

    @patch('qrl.core.p2p.p2pfactory.config', autospec=True)
    def test_add_connection_connection_limit(self, m_config, m_reactor, m_logger):
        """
        When we've reached the connection_limit, add_connection() should refuse to do anything.
        In fact, it should disconnect the P2PProtocol.
        """
        channel_4 = Mock(autospec=P2PProtocol,
                         name='mock Channel 4',
                         peer=IPMetadata('4.4.4.4', 9000))

        m_config.user.max_peers_limit = 3
        m_config.user.max_redundant_connections = 5

        self.assertFalse(self.factory.add_connection(channel_4))

        self.assertEqual(self.factory.num_connections, 3)

    @patch('qrl.core.p2p.p2pfactory.config', autospec=True)
    def test_add_connection_redundant_connection_limit(self, m_config, m_reactor, m_logger):
        """
        When we've reached the max_redundant_connections, then the peer should be disconnected.
        """
        channel_4 = Mock(autospec=P2PProtocol,
                         name='mock Channel 1',
                         peer=IPMetadata('1.1.1.1', 9000))

        m_config.user.max_peers_limit = 5
        m_config.user.max_redundant_connections = 2

        self.assertTrue(self.factory.add_connection(channel_4))
        self.assertFalse(self.factory.add_connection(channel_4))

        self.assertEqual(self.factory.num_connections, 4)

    def test_add_connection_wont_connect_to_itself(self, m_reactor, m_logger):
        """
        P2PFactory should refuse to connect to a P2PProtocol which seems to have our own IP address
        P2PProtocol.host_ip should always be our IP address.
        If add_connection() detects this, then it will rebuild the peer_list, excluding this IP address.
        """
        channel_4 = Mock(autospec=P2PProtocol, name='mock Channel 4',
                         host=IPMetadata('4.4.4.4', config.user.p2p_public_port),
                         peer=IPMetadata('4.4.4.4', config.user.p2p_public_port))
        self.factory._qrl_node.peer_manager.known_peer_addresses = [make_address('1.1.1.1'),
                                                                    make_address('2.2.2.2'),
                                                                    make_address('3.3.3.3'),
                                                                    make_address('4.4.4.4')]
        self.assertFalse(self.factory.add_connection(channel_4))

        self.assertEqual(self.factory.num_connections, 3)
        self.factory._qrl_node.peer_manager.extend_known_peers.assert_called_once_with(
            [make_address('1.1.1.1'), make_address('2.2.2.2'), make_address('3.3.3.3')])

    def test_is_block_present(self, m_reactor, m_logger):
        """
        is_block_present() returns True if block's headerhash is in our blockchain, or if it is known and will be
        processed in the future (POW.future_blocks, OrderedDict())
        """
        self.factory.pow.future_blocks = OrderedDict()
        self.factory._chain_manager._state.get_block.return_value = False
        result = self.factory.is_block_present(b'1234')
        self.assertFalse(result)

        self.factory.pow.future_blocks = OrderedDict({b'1234': 'Some data'})
        result = self.factory.is_block_present(b'1234')
        self.assertTrue(result)

        self.factory.pow.future_blocks = OrderedDict()
        self.factory._chain_manager._state.get_block.return_value = True
        result = self.factory.is_block_present(b'1234')
        self.assertTrue(result)

    def test_connect_peer_already_connected(self, m_reactor, m_logger):
        """
        connect_peer() should not connect to already connected peers.
        We are connected to 1.1.1.1, 2.2.2.2, 3.3.3.3
        :param m_reactor:
        :param m_logger:
        :return:
        """
        self.factory.connect_peer(make_address('1.1.1.1'))
        m_reactor.connectTCP.assert_not_called()

        m_reactor.connectTCP.reset_mock()
        self.factory.connect_peer('1.1.1.1')
        m_reactor.connectTCP.assert_not_called()

    def test_connect_peer(self, m_reactor, m_logger):
        """
        connecting to previously unconnected peers should work.
        """
        self.factory.connect_peer('127.0.0.1:9000')
        m_reactor.connectTCP.assert_called_once()

        m_reactor.connectTCP.reset_mock()
        self.factory.connect_peer('5.5.5.5:9000')
        m_reactor.connectTCP.assert_called_once()

        m_reactor.connectTCP.reset_mock()
        self.factory.connect_peer('5.5.5.5')
        m_reactor.connectTCP.assert_called_once()

    def test_connect_peer_bad_ipport(self, m_reactor, m_logger):
        self.factory.connect_peer('1.1.1.1:65536')
        m_reactor.connectTCP.assert_not_called()

        self.factory.connect_peer('1.1.1.1:')
        m_reactor.connectTCP.assert_not_called()

    def test_monitor_connections(self, m_reactor, m_logger):
        """
        Tries to connect to trusted peers in config.user.peer_list, if they're not already connected.
        Since this function runs periodically, it ensures that the node will always periodically try to connect
        to these preferred peer IPs.

        If it is not connected to any peer, it won't do anything, because its job is to MONITOR, not to connect to
        new peers.
        """
        self.factory.connect_peer = Mock(autospec=P2PFactory.connect_peer)
        with patch('qrl.core.p2p.p2pfactory.config', autospec=True) as m_config:
            m_config.user.peer_list = ['1.1.1.1', '2.2.2.2', '3.3.3.3', '4.4.4.4']
            self.factory.monitor_connections()
            self.factory.connect_peer.assert_called_once_with(make_address('4.4.4.4'))

    def test_monitor_connections_no_peers_connected(self, m_reactor, m_logger):
        self.factory.remove_connection(self.channel_1)
        self.factory.remove_connection(self.channel_2)
        self.factory.remove_connection(self.channel_3)

        self.factory.connect_peer = Mock(autospec=P2PFactory.connect_peer)
        self.factory.monitor_connections()
        self.factory.connect_peer.assert_not_called()

    def test_request_full_message(self, m_reactor, m_logger):
        """
        MessageReceipt is only for Blocks and Transactions, both of which have hashes.
        When you receive the MessageReceipt for them, if you find that you don't have that message corresponding to the
        hash, this is when you call request_full_message()

        If the local node already has the Message, its hash will be in self.master_mr._hash_msg (perhaps it got it
        from another peer). So in that case we check if we requested this hash from any other peer.
        if yes then delete that request and then we go through peer list, who broadcasted this same MR.

        and we request from any one of those peer, and we add that peer into our list, that we already have requested
        from this peer.

        We also make the callLater to the request_full_message, so that in a given interval of time,
        if peer doesn't respond, then we can request to next peer otherwise we can delete all the request, if we
        received it from the peer
        """
        # The local node does not already have this message.
        mrData = qrllegacy_pb2.MRData(type=qrllegacy_pb2.LegacyMessage.SL, hash=b'1234')

        # You can get this Message from channel_1. No, we have not requested this Message from channel_1 yet.
        message_request = MessageRequest()
        message_request.peers_connection_list.append(self.channel_1)
        message_request.already_requested_peers = []
        self.factory.master_mr.requested_hash[b'1234'] = message_request

        self.factory.request_full_message(mrData)

        self.channel_1.send.assert_called_once()

        # We use reactor.callLater() to schedule a call to request_full_message() again in the future.
        # So if this peer doesn't respond, when this function is next called again, it will ignore this peer
        # because of already_requested_peers and use the next one in peers_list.
        m_reactor.callLater.assert_called_once()

    def test_request_full_message_already_requested_this_message_from_another_peer(self, m_reactor, m_logger):
        """
        If we have already requested this Message (from another peer), this function should still go ahead
        and request it from the peer we are currently dealing with.
        """
        # The local node does not already have this message.
        mrData = qrllegacy_pb2.MRData(type=qrllegacy_pb2.LegacyMessage.SL, hash=b'1234')

        # You can get this Message from channel_1. Also, we already requested this Message from channel_2.
        message_request = MessageRequest()
        message_request.peers_connection_list.append(self.channel_1)
        message_request.already_requested_peers = [self.channel_2]
        self.factory.master_mr.requested_hash[b'1234'] = message_request

        self.factory.request_full_message(mrData)

        # But, the code should ignore channel_2 and ask channel_1 anyway.
        self.channel_1.send.assert_called_once()

        # We use reactor.callLater() to schedule a call to request_full_message() again in the future.
        # So if this peer doesn't respond, when this function is next called again, it will ignore this peer
        # because of already_requested_peers and use the next one in peers_list.
        m_reactor.callLater.assert_called_once()

    def test_request_full_message_already_requested_this_message_from_same_peer(self, m_reactor, m_logger):
        # The local node does not already have this message.
        mrData = qrllegacy_pb2.MRData(type=qrllegacy_pb2.LegacyMessage.SL, hash=b'1234')

        # You can get this Message from channel_1 and channel_2. Also, we already requested this Message from channel_1.
        message_request = MessageRequest()
        message_request.peers_connection_list = [self.channel_1, self.channel_2]
        message_request.already_requested_peers = [self.channel_1]
        self.factory.master_mr.requested_hash[b'1234'] = message_request

        self.factory.request_full_message(mrData)

        # We should leave channel_1 alone, but we ask channel_2 for the Message.
        self.channel_1.send.assert_not_called()
        self.channel_2.send.assert_called_once()

        # We use reactor.callLater() to schedule a call to request_full_message() again in the future.
        # So if this peer doesn't respond, when this function is next called again, it will ignore this peer
        # because of already_requested_peers and use the next one in peers_list.
        m_reactor.callLater.assert_called_once()

    def test_request_full_message_we_already_have_this_message(self, m_reactor, m_logger):
        # The local node already has this message!
        mrData = qrllegacy_pb2.MRData(type=qrllegacy_pb2.LegacyMessage.SL, hash=b'1234')
        self.factory.master_mr._hash_msg[b'1234'] = Mock(autospec=Message)
        message_request = MessageRequest()
        message_request.peers_connection_list.append(self.channel_1)
        self.factory.master_mr.requested_hash[b'1234'] = message_request

        self.factory.request_full_message(mrData)

        # Because we already have this message, channel_1 is left alone.
        self.channel_1.send.assert_not_called()
        # Also, this hash should no longer appear in the Master MessageReceipt.
        self.assertIsNone(self.factory.master_mr.requested_hash.get(b'1234'))

    def test_request_full_message_no_peer_could_provide_full_message(self, m_reactor, m_logger):
        """
        If this happens, we should forget about the MessageReceipt and its hash completely.
        Optionally punish peers.
        """
        # The local node does not already have this message.
        mrData = qrllegacy_pb2.MRData(type=qrllegacy_pb2.LegacyMessage.SL, hash=b'1234')

        # No idea where we can get this Message from. We haven't requested this Message from anybody.
        message_request = MessageRequest()
        message_request.peers_connection_list = []
        message_request.already_requested_peers = []
        self.factory.master_mr.requested_hash[b'1234'] = message_request

        self.factory.request_full_message(mrData)

        self.channel_1.send.assert_not_called()
        self.assertIsNone(self.factory.master_mr.requested_hash.get(b'1234'))

        # Now that we've completely forgotten about the MessageReceipt and its hash,
        # the case in test_request_full_message_we_have_already_forgotten_about_this_hash() will happen.

    def test_request_full_message_we_have_already_forgotten_about_this_hash(self, m_reactor, m_logger):
        """
        If we couldn't download the full Message from any peer, then we would've forgotten about this MessageReceipt
        and its hash.
        The next time request_full_message() is called, we find that we have forgotten about the MR and thus do nothing.
        """
        mrData = qrllegacy_pb2.MRData(type=qrllegacy_pb2.LegacyMessage.SL, hash=b'1234')

        self.factory.request_full_message(mrData)

        self.channel_1.send.assert_not_called()
        m_reactor.callLater.assert_not_called()

    def test_broadcast(self, m_reactor, m_logger):
        # broadcast msg_type is qrllegacy_pb2.LegacyMessage
        self.factory.broadcast(qrllegacy_pb2.LegacyMessage.TX, b'1234')
        self.channel_1.send.assert_called_once()
        self.channel_2.send.assert_called_once()
        self.channel_3.send.assert_called_once()

    def test_broadcast_does_not_broadcast_to_peers_known_to_have_the_mr(self, m_reactor, m_logger):
        message_request = MessageRequest()
        message_request.peers_connection_list = [self.channel_2, self.channel_3]
        self.factory.master_mr.requested_hash[b'1234'] = message_request
        self.factory.broadcast(qrllegacy_pb2.LegacyMessage.SL, b'1234')
        self.channel_1.send.assert_called_once()
        self.channel_2.send.assert_not_called()
        self.channel_3.send.assert_not_called()

    def test_broadcast_tx(self, m_reactor, m_logger):
        # broadcast_tx() should handle all Transaction Types
        self.factory.broadcast_tx(MessageTransaction())
        self.factory.broadcast_tx(TransferTransaction())
        self.factory.broadcast_tx(TokenTransaction())
        self.factory.broadcast_tx(TransferTokenTransaction())
        self.factory.broadcast_tx(SlaveTransaction())
        with self.assertRaises(ValueError):
            m_tx = Mock(autospec=TransferTransaction, txhash=bhstr2bin('deadbeef'))
            self.factory.broadcast_tx(m_tx)

    def test_broadcast_block(self, m_reactor, m_logger):
        """
        Not much to test here, other than that it works
        """
        tx = Mock(autospec=TransferTransaction, public_key=b'1234')
        b = Mock(autospec=Block, transactions=[tx], headerhash=bhstr2bin('1234'), block_number=1,
                 prev_headerhash=bhstr2bin('deadbeef'))
        self.factory.broadcast_block(b)

    def test_get_random_peer(self, m_reactor, m_logger):
        """
        1. Figures out max cumulative difficulty from what it knows about other peers
        2. Fills best_connection_ids with peers that have the max_cumulative_difficulty
        3. selected_peer_connections is simply a list of P2PProtocols that correspond to the IPs in best_connection_ids
        """
        self.factory.update_peer_blockheight(make_address('1.1.1.1'), 5, b'555555', StringToUInt256('5'))
        result = self.factory.get_random_peer()
        self.assertEqual(result, self.channel_1)

        # Even though block_number is the same, it will pick channel_2 because the difficulty is higher.
        self.factory.update_peer_blockheight(make_address('2.2.2.2'), 5, b'555555', StringToUInt256('6'))
        result = self.factory.get_random_peer()
        self.assertEqual(result, self.channel_2)

        # Now it should either return channel_2 or channel_3
        self.factory.update_peer_blockheight(make_address('3.3.3.3'), 5, b'555555', StringToUInt256('6'))
        result = self.factory.get_random_peer()
        self.assertIn(result, [self.channel_2, self.channel_3])

    def test_get_random_peer_no_peers(self, m_reactor, m_logger):
        """
        If all peers have max cumulative difficulty 0, why bother returning them?
        """
        self.factory.update_peer_blockheight('1.1.1.1:9000', 5, b'555555', StringToUInt256('0'))
        self.factory.update_peer_blockheight('2.2.2.2:9000', 5, b'555555', StringToUInt256('0'))
        self.factory.update_peer_blockheight('3.3.3.3:9000', 5, b'555555', StringToUInt256('0'))

        result = self.factory.get_random_peer()
        self.assertIsNone(result)

    def test_request_peer_blockheight(self, m_reactor, m_logger):
        self.factory.request_peer_blockheight()
        self.channel_1.send.assert_called_once()
        self.channel_2.send.assert_called_once()
        self.channel_3.send.assert_called_once()

        # 0 means that we are asking for the blockheight, not telling them ours.
        self.assertEqual(0, self.channel_1.send.call_args[0][0].bhData.block_number)

    def test_add_unprocessed_txn(self, m_reactor, m_logger):
        """
        This function is very simple. It simply tries to add the tx to the tx_pool.
        If that works, then it tries to start the txn_processor if it isn't already started.
        If the tx_pool can't be added to, there's nothing we can do and return False.
        """
        self.factory._chain_manager = Mock(autospec=ChainManager)
        self.factory._chain_manager.tx_pool.update_pending_tx_pool.return_value = True
        m_tx = Mock(autospec=TransferTransaction)
        m_tx.fee = 0

        self.assertFalse(self.factory._txn_processor_running)
        # It worked, and a TxnProcessor was started
        result = self.factory.add_unprocessed_txn(m_tx, '1.1.1.1')
        self.assertTrue(result)
        self.assertTrue(self.factory._txn_processor_running)

    def test_add_unprocessed_txn_txnprocessor_already_running(self, m_reactor, m_logger):
        """
        If the txnprocessor is already running, return True as usual.
        """
        self.factory._chain_manager = Mock(autospec=ChainManager)
        self.factory._chain_manager.tx_pool.update_pending_tx_pool.return_value = True
        self.factory._txn_processor_running = True
        m_tx = Mock(autospec=TransferTransaction)
        m_tx.fee = 0

        # It worked
        result = self.factory.add_unprocessed_txn(m_tx, '1.1.1.1')
        self.assertTrue(result)

    def test_add_unprocessed_txn_add_to_txpool_failed(self, m_reactor, m_logger):
        """
        If we couldn't add to txpool, don't start the txnprocessor. And return False.
        """
        self.factory._chain_manager = Mock(autospec=ChainManager)
        self.factory._chain_manager.tx_pool.update_pending_tx_pool.return_value = False
        m_tx = Mock(autospec=TransferTransaction)
        m_tx.fee = 0

        result = self.factory.add_unprocessed_txn(m_tx, '1.1.1.1')
        self.assertFalse(result)


@patch('qrl.core.misc.ntp.getTime', new=replacement_getTime)
@patch('qrl.core.p2p.p2pfactory.logger', autospec=logger, name="Mock Logger")
@patch('qrl.core.p2p.p2pfactory.reactor', autospec=reactor, name="My Mock Reactor")
class TestP2PFactoryCompareAndSync(TestCase):
    def setUp(self):
        def replacement_get_block_by_number(i):
            return self.blocks[i - 1]

        port = '9000'
        self.m_qrlnode = Mock(autospec=QRLNode, name='Fake QRLNode')
        self.factory = P2PFactory(chain_manager=ChainManager(state=Mock(autospec=State)), sync_state=None,
                                  qrl_node=self.m_qrlnode)

        self.factory.peer_fetch_block = Mock(autospec=P2PFactory.peer_fetch_block)
        self.factory._chain_manager.get_block_by_number = replacement_get_block_by_number

        self.channel_1 = Mock(autospec=P2PProtocol, name='mock Channel 1', peer_ip='1.1.1.1', peer_port=port)
        self.factory.add_connection(self.channel_1)

        self.blocks = [
            Mock(autospec=Block, name='0th Block', block_number=1, headerhash=bhstr2bin('123456')),
            Mock(autospec=Block, name='1st Block', block_number=2, headerhash=bhstr2bin('7890ab')),
            Mock(autospec=Block, name='2nd Block', block_number=3, headerhash=bhstr2bin('deadbeef'))
        ]
        self.node_header_hash = qrl_pb2.NodeHeaderHash(
            block_number=1,
            headerhashes=[bhstr2bin('123456'), bhstr2bin('7890ab'), bhstr2bin('deadbeef')]
        )

    def test_compare_and_sync_no_fork(self, m_reactor, m_logger):
        """
        compare_and_sync() looks for forks by comparing headerhashes in the peer's NodeHeaderHash and the node's blocks.
        There is no fork
        """

        # If it finds out we're on the last block of the NodeHeaderHash list, then it doesn't ask the peer for more blocks.
        self.factory._chain_manager._last_block = self.blocks[-1]
        self.factory.compare_and_sync(self.channel_1, self.node_header_hash)
        self.factory.peer_fetch_block.assert_not_called()

        # If we are in the middle of the NodeHeaderHash list, it asks for the next block.
        self.factory._chain_manager._last_block = self.blocks[0]
        self.factory.compare_and_sync(self.channel_1, self.node_header_hash)
        self.factory.peer_fetch_block.assert_called_once()

    def test_compare_and_sync_detect_fork(self, m_reactor, m_logger):
        """
        There is a fork at block_number 2.
        peer_fetch_block() should be called.
        """
        # Mess with the expected headerhash value
        self.node_header_hash = qrl_pb2.NodeHeaderHash(
            block_number=1,
            headerhashes=[bhstr2bin('123456'), bhstr2bin('000000'), bhstr2bin('deadbeef')]
        )
        # Set our state as being on the 2nd block of the NodeHeaderHash
        self.factory._chain_manager._last_block = self.blocks[1]

        self.factory.compare_and_sync(self.channel_1, self.node_header_hash)

        self.factory.peer_fetch_block.assert_called_once()

    def test_compare_and_sync_nonsense_block_number(self, m_reactor, m_logger):
        """
        The NodeHeaderHash has a block_number, which indicates the block_number of the the first headerhash.
        To be processing a block with a block_number less than NodeHeaderHash.block_number makes no sense.
        """
        # Mess with Block 2's block_number
        self.blocks[1].block_number = 0

        # Set our state as being on the 2nd block of the NodeHeaderHash
        self.factory._chain_manager.last_block.return_value = self.blocks[1]
        self.factory.compare_and_sync(self.channel_1, self.node_header_hash)

        self.factory.peer_fetch_block.assert_not_called()


@patch('qrl.core.misc.ntp.getTime', new=replacement_getTime)
@patch('qrl.core.p2p.p2pfactory.logger', autospec=logger, name="Mock Logger")
@patch('qrl.core.p2p.p2pfactory.reactor', autospec=reactor, name="My Mock Reactor")
class TestP2PFactoryPeerFetchBlock(TestCase):
    def setUp(self):
        port = '9000'
        self.m_qrlnode = Mock(autospec=QRLNode, name='Fake QRLNode')
        self.channel_1 = Mock(autospec=P2PProtocol, name='mock Channel 1', peer_ip='1.1.1.1', peer_port=port)
        self.channel_2 = Mock(autospec=P2PProtocol, name='mock Channel 2', peer_ip='2.2.2.2', peer_port=port)
        self.channel_3 = Mock(autospec=P2PProtocol, name='mock Channel 3', peer_ip='3.3.3.3', peer_port=port)

        self.factory = P2PFactory(chain_manager=ChainManager(state=Mock(autospec=State)), sync_state=None,
                                  qrl_node=self.m_qrlnode)
        self.factory.pow = Mock(autospec=POW)

        self.m_qrlnode.is_banned.return_value = False
        self.factory.add_connection(self.channel_1)
        self.factory.add_connection(self.channel_2)
        self.factory.add_connection(self.channel_3)

        self.factory._target_channel = self.channel_1
        self.factory.is_syncing_finished = Mock(return_value=False, autospec=P2PFactory.is_syncing_finished)
        self.factory._target_node_header_hash = qrl_pb2.NodeHeaderHash(
            block_number=1,
            headerhashes=[bhstr2bin('123456'), bhstr2bin('deadbeef'), bhstr2bin('abcdef')]
        )
        self.factory._last_requested_block_number = 1

    def test_peer_fetch_block_we_dont_already_have_the_block(self, m_reactor, m_logger):
        """
        Given a peer's NodeHeaderHash (inventory), peer_fetch_block() tries to find a corresponding block in the node's
        chain. If it can't find it then it sends a fetch_block Message (P2PProtocol.send_fetch_block())
        """
        self.factory._chain_manager._state.get_block.return_value = None

        self.factory.peer_fetch_block()

        self.channel_1.send_fetch_block.assert_called_once_with(1)

    def test_peer_fetch_block_we_already_have_the_block(self, m_reactor, m_logger):
        """
        If peer_fetch_block() finds a corresponding block in the node's chain, then it keeps asking the node's local
        chain for newer blocks until
        1. the node doesn't have any newer blocks OR
        2. we've reached the end of the peer's NodeHeaderHash.
        Then it will ask the peer for the next block after that.
        """
        self.factory._chain_manager._state.get_block.return_value = Block()

        self.factory.peer_fetch_block()

        self.assertEqual(self.factory._chain_manager._state.get_block.call_count, 3)
        self.channel_1.send_fetch_block.assert_called_once_with(3)

    def test_peer_fetch_block_we_are_synced(self, m_reactor, m_logger):
        """
        If is_syncing_finished() is True, then let's not ask for more blocks.
        """
        self.factory._chain_manager._state.get_block.return_value = Block()
        self.factory.is_syncing_finished.return_value = True

        self.factory.peer_fetch_block()

        self.channel_1.send_fetch_block.assert_not_called()

    def test_peer_fetch_block_block_not_found_and_retry(self, m_reactor, m_logger):
        """
        If we can't find the blocks that the Peer's NodeHeaderHash indicated, and retry >= 5, ban the peer.
        Assume we are synced.
        """
        self.factory._chain_manager._state.get_block.return_value = None

        self.factory.peer_fetch_block(retry=5)

        self.m_qrlnode.peer_manager.ban_channel.assert_called_once_with(self.channel_1)
        self.factory.is_syncing_finished.assert_called_once_with(force_finish=True)


@patch('qrl.core.misc.ntp.getTime', new=replacement_getTime)
@patch('qrl.core.p2p.p2pfactory.logger', autospec=logger, name="Mock Logger")
@patch('qrl.core.p2p.p2pfactory.reactor', autospec=reactor, name="My Mock Reactor")
class TestP2PFactoryBlockReceived(TestCase):
    def setUp(self):
        port = '9000'
        self.m_qrlnode = Mock(autospec=QRLNode, name='Fake QRLNode')
        self.channel_1 = Mock(autospec=P2PProtocol, name='mock Channel 1', peer_ip='1.1.1.1', peer_port=port)
        self.channel_2 = Mock(autospec=P2PProtocol, name='mock Channel 2', peer_ip='2.2.2.2', peer_port=port)
        self.channel_3 = Mock(autospec=P2PProtocol, name='mock Channel 3', peer_ip='3.3.3.3', peer_port=port)

        self.factory = P2PFactory(chain_manager=Mock(autospec=ChainManager), sync_state=None, qrl_node=self.m_qrlnode)
        self.factory.pow = Mock(autospec=POW)

        self.m_qrlnode.is_banned.return_value = False
        self.factory.add_connection(self.channel_1)
        self.factory.add_connection(self.channel_2)
        self.factory.add_connection(self.channel_3)

        # The peer has two blocks. It has sent a list of their hashes, and block_number indicates that the first hash
        # in the list is for block_number 1.
        self.factory._target_node_header_hash = qrl_pb2.NodeHeaderHash(
            block_number=1,
            headerhashes=[bhstr2bin('123456'), bhstr2bin('deadbeef')]
        )

        # This mocking ensures that the Block gets added to the Chain, and that the next block is requested.
        self.factory._target_peer = self.channel_1
        self.factory._last_requested_block_number = 1
        self.factory._chain_manager.add_block.return_value = True
        self.factory.peer_fetch_block = Mock(autospec=P2PFactory.peer_fetch_block)

    def tearDown(self):
        self.factory.remove_connection(self.channel_1)
        self.factory.remove_connection(self.channel_2)
        self.factory.remove_connection(self.channel_3)

    def test_block_received(self, m_reactor, m_logger):
        """
        P2PChainManager.handle_push_block() simply creates the Block from the incoming Protobuf data, and throws an
        error if it can't.
        But this function is the one that actually:
        1. checks that the block came from the peer we were talking to
        2. checks that the block is the one we requested
        3. validates the block
        4. requests the next block (since we are in a syncing scenario)
        """
        m_reactor.download_monitor = Mock()
        # The peer has sent the whole block that corresponds to block_number 1.
        block = Mock(autospec=Block, block_number=1, headerhash=bhstr2bin('123456'))
        block.validate.return_value = True

        self.factory._target_channel = self.channel_1
        self.factory.block_received(self.channel_1, block)

        self.factory._chain_manager.add_block.assert_called_once()
        # After all this, the node should ask for the next block.
        self.factory.peer_fetch_block.assert_called_once()

        # After the peer sends the last block, we should not ask for any more blocks.
        self.factory._chain_manager.add_block.reset_mock()
        self.factory.peer_fetch_block.reset_mock()
        block_2 = Mock(autospec=Block, block_number=2, headerhash=bhstr2bin('deadbeef'))
        block.validate.return_value = True

        self.factory.block_received(self.channel_1, block_2)

        self.factory._chain_manager.add_block.assert_called_once()
        self.assertIsNone(self.factory._target_channel)
        self.factory.peer_fetch_block.assert_not_called()

    def test_block_received_suspend_mining_when_we_just_updated_chain(self, m_reactor, m_logger):
        """
        If we've just updated the chain with a new block, start mining a bit later, not now.
        """
        m_reactor.download_monitor = Mock()
        self.factory._chain_manager.last_block.headerhash = bhstr2bin('123456')
        block = Mock(autospec=Block, block_number=1, headerhash=bhstr2bin('123456'))
        self.factory.block_received(self.channel_1, block)
        self.assertNotEqual(self.factory.pow.suspend_mining_timestamp, 0)

    def test_block_received_wrong_peer(self, m_reactor, m_logger):
        """
        If the incoming Block didn't come from the peer we were talking to, don't process (ignore) it.
        """
        m_reactor.download_monitor = Mock()
        block = Mock(autospec=Block, block_number=1, headerhash=bhstr2bin('123456'))
        block.validate.return_value = True

        self.factory._target_channel = self.channel_2

        self.factory.block_received(self.channel_1, block)

        self.factory._chain_manager.add_block.assert_not_called()
        self.factory.peer_fetch_block.assert_not_called()

    def test_block_received_wrong_block_idx(self, m_reactor, m_logger):
        """
        If the incoming block_number is not the one we requested, don't process it.
        """
        m_reactor.download_monitor = Mock()
        block = Mock(autospec=Block, block_number=2, headerhash=bhstr2bin('123456'))
        block.validate.return_value = True

        self.factory.block_received(self.channel_1, block)

        self.factory._chain_manager.add_block.assert_not_called()
        self.factory.peer_fetch_block.assert_not_called()

    def test_block_received_wrong_block_headerhash(self, m_reactor, m_logger):
        """
        The node should already know which block_number should have which headerhash.
        """
        m_reactor.download_monitor = Mock()
        block = Mock(autospec=Block, block_number=1, headerhash=bhstr2bin('deadbeef'))
        block.validate.return_value = True

        self.factory.block_received(self.channel_1, block)

        self.factory._chain_manager.add_block.assert_not_called()
        self.factory.peer_fetch_block.assert_not_called()

    def test_block_received_block_fails_validation(self, m_reactor, m_logger):
        """
        Somehow the block fails to validate. Then it shouldn't be added to the chain.
        """
        m_reactor.download_monitor = Mock()
        block = Mock(autospec=Block, block_number=1, headerhash=bhstr2bin('123456'))
        block.validate.return_value = False

        self.factory.block_received(self.channel_1, block)

        self.factory._chain_manager.add_block.assert_not_called()
        self.factory.peer_fetch_block.assert_not_called()

    def test_block_received_adding_block_fails(self, m_reactor, m_logger):
        """
        If the block couldn't be added to the chain for any reason, don't ask for the next block.
        In fact, don't even run is_syncing_finished(), for it might mess up our current state.
        """
        m_reactor.download_monitor = Mock()
        block = Mock(autospec=Block, block_number=1, headerhash=bhstr2bin('123456'))
        block.validate.return_value = True
        self.factory._chain_manager.add_block.return_value = False
        self.factory.is_syncing_finished = Mock(autospec=P2PFactory.is_syncing_finished)

        self.factory.block_received(self.channel_1, block)

        self.factory.peer_fetch_block.assert_not_called()
        self.factory.is_syncing_finished.assert_not_called()



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
import contextlib
from unittest import TestCase

from qrl.core import config
from qrl.core.misc import logger
from qrl.generated import qrl_pb2
from qrl.services.PublicAPIService import PublicAPIService
from tests.blockchain.MockedBlockchain import MockedBlockchain

logger.initialize_default()


@contextlib.contextmanager
def set_timeseries_size(new_size):
    old_value = config.dev.block_timeseries_size
    try:
        config.dev.block_timeseries_size = new_size
        yield
    finally:
        config.dev.block_timeseries_size = old_value


class TestPublicAPI(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestPublicAPI, self).__init__(*args, **kwargs)

    def test_getStats(self):
        number_blocks = 1
        with MockedBlockchain.create(number_blocks) as mock_blockchain:
            with set_timeseries_size(10):
                service = PublicAPIService(mock_blockchain.qrlnode)

                for i in range(number_blocks + 1, 50):
                    mock_blockchain.add_new_block()

                    request = qrl_pb2.GetStatsReq()
                    request.include_timeseries = True

                    stats = service.GetStats(request=request, context=None)

                    self.assertEqual(min(10, i + 1), len(stats.block_timeseries))

    def getStats_forking(self, timeseries_size):
        number_blocks = 10
        with MockedBlockchain.create(number_blocks) as mock_blockchain:
            with set_timeseries_size(timeseries_size):
                service = PublicAPIService(mock_blockchain.qrlnode)

                request = qrl_pb2.GetStatsReq(include_timeseries=True)
                stats = service.GetStats(request=request, context=None)
                self.assertEqual(timeseries_size, len(stats.block_timeseries))

                # Fork at node 9 with 2 blocks (b)
                block_9 = mock_blockchain.qrlnode.get_block_from_index(9)
                block_10 = mock_blockchain.qrlnode.get_block_last()
                block_10b = mock_blockchain.create_block(block_9.headerhash)
                mock_blockchain.add_block(block_10b)
                block_11b = mock_blockchain.create_block(block_10b.headerhash)
                mock_blockchain.add_block(block_11b)

                # Get time series
                request = qrl_pb2.GetStatsReq(include_timeseries=True)
                stats = service.GetStats(request=request, context=None)
                self.assertEqual(timeseries_size, len(stats.block_timeseries))

                # check stats have moved to the correct branch
                self.assertEqual(block_11b.headerhash, stats.block_timeseries[-1].header_hash)

                # Add back to the original chain (a)
                block_11 = mock_blockchain.create_block(block_10.headerhash)
                mock_blockchain.add_block(block_11)
                block_12 = mock_blockchain.create_block(block_11.headerhash)
                mock_blockchain.add_block(block_12)

                # Get time series
                request = qrl_pb2.GetStatsReq(include_timeseries=True)
                stats = service.GetStats(request=request, context=None)
                self.assertEqual(timeseries_size, len(stats.block_timeseries))
                # check stats have moved back to the correct branch
                self.assertEqual(block_12.headerhash, stats.block_timeseries[-1].header_hash)

    def test_getStats_forking1(self):
        # FIXME: Parametrized testing is different in nose/pytest and that brings incompatibilities
        self.getStats_forking(timeseries_size=1)

    def test_getStats_forking5(self):
        self.getStats_forking(timeseries_size=5)

    def test_getStats_forking10(self):
        self.getStats_forking(timeseries_size=10)



from unittest import TestCase

from qrl.core.misc import logger
from qrl.core.txs.Transaction import Transaction

logger.initialize_default()


class TestTransaction(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestTransaction, self).__init__(*args, **kwargs)

    def test_calc_allowed_decimals(self):
        decimal = Transaction.calc_allowed_decimals(10000000000000000000)
        self.assertEqual(decimal, 0)

        decimal = Transaction.calc_allowed_decimals(1)
        self.assertEqual(decimal, 19)

        decimal = Transaction.calc_allowed_decimals(2)
        self.assertEqual(decimal, 18)



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
import threading

from qrl.core.Singleton import Singleton
from pyqryptonight import pyqryptonight


class Qryptonight(object, metaclass=Singleton):

    def __init__(self):
        self.lock = threading.Lock()
        self._qn = pyqryptonight.Qryptonight()

    def hash(self, blob):
        with self.lock:
            return bytes(self._qn.hash(blob))
